# External Entities

This module allows you to connect to and use datasets from external sources in
your Drupal website. While content lives externally, Drupal sees those remote
entities as internal. This makes it possible to alter the field displays, add
references, comments, path aliases, share buttons, web-shop products and more.

For a full description of the module, visit the 
[project page](https://www.drupal.org/project/external_entities).

Submit bug reports and feature suggestions, or track changes in the
[issue queue](https://www.drupal.org/project/issues/external_entities).


## Table of contents

- Requirements
- Installation
- Configuration
- How it works / How to
- Recommended modules
- v2 to v3 changes
- Maintainers


## Requirements

This module requires no modules outside of Drupal core but will require the
JsonPath PHP library ([galbar/jsonpath](https://github.com/Galbar/JsonPath-PHP))
which is automatically installed by composer.


## Installation

Install as you would normally install a contributed Drupal module. For further
information, see
[Installing Drupal Modules](https://www.drupal.org/docs/extending-drupal/installing-drupal-modules).


## Configuration

1. Enable the module at Administration > Extend.
2. To enable integration of external entities, create a new external entity on
   admin/structure/external-entity-types.
3. Configure the external source, add (Drupal) fields, save and edit the
   external entity again to configure field mappings.

For more information, see the documentation page about
[External Entities](https://www.drupal.org/docs/contributed-modules/external-entities).


## How it works / How to

### Multiple storage client aggregation

#### Aggregation modes

When combining data from multiple storage client sources into one entity, three
use cases are possible:

1. each data source adds properties to the entities of the previous source(s),
  either because the entity data from both sources share the same identifier or
  the value of a specified field of the new data source entity matches the
  value of a specified field from the previous sources (ie. "join").

2. each data source brings its own set of entities. Therefore, entity
  identifiers should all differ from one data source to another. If it is not
  the case, data source entity identifiers could be prefixed by xnttmulti
  using a "virtual group prefix" on a data source-basis making each identifier
  unique (ie. source A as an entity with id '001', source B as a different
  entity that also uses the id '001'; xnttmulti can virtuall prefix A ids with
  'A' and B ids with 'B' which would lead to 2 distinct identifiers: 'A001' and
  'B001' which clearly identifies each source).

3. a mix of the 2 previous cases.

This modules offers different approaches to deal with those 3 cases but with
some restrictions. Those restrictions are here to solve some technical aspects:
 - being able to efficiently count and filter external entities without needing
   to query all the data from all the sources and process all of them each time.
 - being able to clearly identify the corresponding element in each data source
   given an identifier without any ambiguity.

For the first case, this module assumes the first data source holds ALL the
entity identifiers and can be used for pagging management. If an identifier is
not present in the first data source but in one (or more) of the next data
sources, the corresponding entity will be discarded. For filters, only
reference sources are used (at the moment).

For the second and third cases, a "group prefix" must be provided to every set
of data in order to discriminate data source according to the given identifiers.
For instance, if we have 2 data sources A and B, we need to know that the
identifier '001' corresponds to an entity stored in A or B. Therefore, either
one of the 2 sources uses a prefix like '0' (or '00') and can be clearly
identified (which mich be necessary when creating a new entity instance to avoid
duplicating it in the 2 sources) or the sources have been assigned a virtual
group prefix that will not be stored (nor known by the data sources) but only
used by xnttmulti to discreminate between sources. We call it a "group prefix"
because a same prefix could be used by multiple data sources in order to
aggregate them together like in the first use case.

The first data source of each group will be considered as the group "reference"
and processed just like in the first use case for aggregation and pagging
management. When group prefixes are used, any entity from data sources with or
without group prefix will only be considered and aggregated if its identifier or
join field value matches an existing entity from previous sources. Group
prefixes may be virtual (stripped and not comunicated to the data source) or
physical (part of the identifier provided by the data source) but all group
prefixes must be distinct (no overlap): you can't use the prefix 'AB' on one
data source and the prefix 'ABC' on another for a same external entity type
using xnttmulti since 'AB' is also a part of the prefix 'ABC'.

There is a special case when the first(s) data source(s) is (are) not using a
group prefix while other sources have one. Then, xnttmulti will use the first
use case behavior and only use the next specified group as filters on each data
source that has a group.

It is possible to specify more than one group prefix to each data source by
separating them with a semicolon. It might be conveinent when some data sources
are common to other distinct sources. For instance a data source A and a data
source B may contain distinct entities while a data source C may contain
additional data for both; then A would use the prefix group "A", B the prefix
group "B" and C would use "A;B".

If a data source uses multiple *virtual* prefixes (ie. not part of the "real"
identifier), each of its entities will be duplicated for each prefix. For
instance, if a data source C contains an entity '001' and uses the virtual group
prefixes "A;B", xnttmulti will consider that entity '001' has 2 copies: 'A001'
and 'B001' (it virtually doubles the number of entities).

When an entity is aggregated from several data sources (same identifier or with
join), it is possible to choose the way to merge the last data source:
- either without override and only new fields are added
- or with override and any field from the last source will replace previous
  existing values, except for the entity identifier field or its join fields
- or as a sub-object which will be hold in a new user-defined field name not
  existing in the sources.

#### Joins

Sometimes, the identifier field has different names on different sources,
sometimes a different field should be used to join a record from a data source
to the corresponding record in another "joined" source.

To manages thoses cases, this module offers 2 setting fields and different
behavior depending on the way they are filled; the first field is "Previous
storage clients field name that provides the identifier value for this storage
client (join field)", which we will call "join field", and the field "Source
(id) field to use to join data", which we will call "source id" that correspond
to a field of the joined source that will be used as id field for that source.

If both "join id" and "source id" are left empty, xnttmulti assumes the default
identifier field must be available on the joined data source and will match it.
If both "join id" and "source id" are filled, xnttmulti will use the field value
of "join id" field from the previous sources as the identifier value to
use to match "source id" field in the new data source.
If "join id" is filled but "source id" is empty, xnttmulti assumes "source id"
field is the same as the one specified for "join id" and proceed just like when
the "join id" and "source id" are filled.
If "join id" is empty but "source id" is filled, xnttmulti will match the
default identifier provided by previous sources with "source id" field against
the new data source (and it will apply prefix virtualization if needed).
Note: when "join id" is filled, the group prefix will not be used to filter data
on the new source; it will only be used to know if the source should be
aggregated to a given group.

### Saving

- non-mapped source field values are left "as they are".
- Drupal fields mapped to constant values are not saved.
- Multiple values for a given source field are replaced by the Drupal values.
- Mapped Drupal fields that have empty values replace/clear source values.
- Drupal fields mapped to non-existing source fields are saved (if the source
  storage takes them into account).
- Source associative arrays which values are mapped to a Drupal field (with
  multiple values) but keys are not mapped (it can depend on the field mapper
  used) will result on original key-value pairs being maintained while new
  values are added using regular numeric indexation. This would result in
  duplicate values (with different keys).

### Creating external entity types programmatically

It is possible to create external entities programmatically with constraints.
See companion example modules in the modules directory for examples.
You can either create external entity types using YAML config files in your
module "config/install/" directory or create them by code in your module install
file (in hook_install()).

The constraints should be defined your module install file "hook_install()"
implementation as follow. Don't forget to also clear those changes in your
"hook_uninstall()" implementation.
```
  $locked = \Drupal::state()->get('external_entities.type.locked');
  // Setup constraints on the external entity type 'xnttex1':
  $locked['xnttex1'] = [
    // To disable completly type editing.
    'lock_edit' => FALSE,
    // To disable type removal.
    'lock_delete' => TRUE,
    // To only disable type field editing (add/remove but not display).
    'lock_fields' => TRUE,
    // To disable type field editing and display changes.
    'lock_display' => FALSE,
    // To disable type annotation changes.
    'lock_annotations' => TRUE,
    // To limit number of type field mappers - useless for the moment.
    'lock_field_mapper_count' => 1,
    'lock_field_mappers' => [
      [
        // To restrict type field mapper selection.
        'allow_plugins' => [
          'simple' => 'simple',
        ],
        // To disable type field mapper config changes.
        'lock_config' => FALSE,
        // To completely hide type field mapper config.
        'hide_config' => FALSE,
      ],
    ],
    'lock_data_aggregator' => [
    // To restrict type data aggregator selection.
      'allow_plugins' => [
        'group' => 'group',
      ],
      // To disable type data aggregator config changes.
      'lock_config' => FALSE,
      // To completely hide type data aggregator config.
      'hide_config' => FALSE,
    ],
    // To limit number of type storage clients.
    'lock_storage_client_count' => 1,
    'lock_storage_clients' => [
      [
        // To restrict type storage client selection.
        'allow_plugins' => [
          'rest' => 'rest',
        ],
        // To disable type storage client config changes.
        'lock_config' => TRUE,
        // To completely hide type storage client config.
        'hide_config' => FALSE,
      ],
    ],
  ];
  Drupal::state()->set('external_entities.type.locked', $locked);
```


## Recommended modules

- [Search API](https://www.drupal.org/project/search_api), when enabled
it is possible to make overviews of external entities.

- [External Entity Manager](https://www.drupal.org/project/xnttmanager), which
allows to inspect available external fields and their mapping, to synchronize
local data with external data and auto-generate associated annotation entities.

- [External Entity Files storage](https://www.drupal.org/project/xnttfiles), a
file storage client plugin which also allows to map Drupal File and Image fields
to external URIs (URLs or stream paths).

- [External Entities Database storage](https://www.drupal.org/project/xnttdb),
an external database storage client plugin which allows use external databases
(both mySQL and Postgres, regardless what Drupal is using) as external data
sources.

- [External Entities Views](https://www.drupal.org/project/xnttviews),
an experimental plugin that enables the use of external entities with Views.

Find more contributed modules at the
[External Entities plugins page](https://www.drupal.org/docs/contributed-modules/external-entities/external-entities-plugins)


## v2 to v3 changes

v3 has been introduced to allow major External Entities upgrades (through
breaking changes) while keeping v2 to let previous system continue working
without being required to upgrade. The v3 changes come from raised issues,
improvements and new feature requests, after several years of use of External
Entities in various contexts.

List of *BREAKING CHANGES* between External Entities v2 and v3:
- Global changes:
  - Type hinting has been added to v3 wherever possible.
  - ExternalEntityInterface and ExternalEntityTypeInterface have been moved to
    the Drupal\external_entities\Entity namespace to regroup them with their
    respective implementations.
  - Storage class name simplification: "ExternalEntityStorageClient*" become
    "StorageClient*".
  - Storage client plugin type name changed from "ExternalEntityStorageClient"
    to "StorageClient".
- Field Mapper
  - Removed unnecessary abstraction layers for field mappers:
    - "ExpressionFieldMapperInterface" has been merged to
      "FieldMapperInterface".
    - "ConfigurableExpressionFieldMapperBase" and "ExpressionFieldMapperBase"
      have beend merge to "FieldMapperBase".
    Those layers came from a discussion when introducing field mapper plugins:
    https://www.drupal.org/project/external_entities/issues/3079012#comment-13437754
    Field mapper plugins should now use "FieldMapperBase" class as base.
    The rational behind this change is that those layers were adding complexity
    to the code since some methods of "ExpressionFieldMapperInterface" were
    required at some places where "FieldMapperInterface" was used. It would be
    hard to choose where to use one interface or the other while only one is
    used by plugins and current external entities implementation forces the use
    of plugins for field mappers. If one day, someone would create external
    entities type from code and want to force the use of a non-plugin field
    mapper, either this person would have to turn it into a plugin (which would
    not be so complicated) or raise an issue with explanations on the use case.
  - "createRawDataFromEntityValues" has an additional parameter
    $original_values that should be the source array which will be overriden
    by the new values.
  - "getFieldPropertyMapping" has been removed and can be replaced by
    "getFieldMapping" as the possible return types would be the same now.
  - "getConstantMappingPrefix" has been removed as the mapping type (ie. for
    instance FieldMapperInterface::MAPPING_CONSTANT) is now provided with the
    mapping string when using "getFieldMapping".
  - "extractFieldValuesFromRawData" and "extractFieldPropertyValuesFromRawData"
    are now part of the interface and made public but are using the string field
    name as first argument instead of the field definition.
  - "getMappableFields" stores mappable fields in class member "mappableFields".
- REST client:
  - Method getListQueryParameters() does not have "$start" and "$length"
    parameters. It does not handle paging anymore as well: paging is managed by
    a modified getPagingQueryParameters() method.
  - Method getPagingQueryParameters() does not return an array of paging
    parameters but an array of paging parameters for multiple queries to map
    client paging to server paging (with different page sizes).
  - The REST client does not save data using "application/x-www-form-urlencoded"
    data by default anymore. It now uses the REST endpoint encoding type set by
    the "Endpoint data encoding format" setting.

List of v3 new features:
- Global:
  - New method to set and retrieve source raw data from an external entity:
    ´setSourceRawData(array)´ and ´getSourceRawData()´. When external entities
    are saved, source fields that were not mapped to Drupal fields are kept.
  - Plugins (Storage Clients and Field Mappers) can now rely on the method
    submitConfigurationForm() rather than validateConfigurationForm() to set
    their configuration changes while still use validateConfigurationForm() to
    validate/invalidate form submission.
  - Added support for debugging external entities (loading, mapping, saving).
  - Both field mapper and storage clients settings can have administrative notes
    for an easier maintenance (document specific setting reasons, other possible
    options not used, etc.).
  - External Entity Types can be created programatically or by config, with or
    without editing restrictions and are highly customizable (list builder,
    forms, etc.).
  - Plugins using PluginFormTrait now benefit from the new method
    ::buildAjaxParentSubForm() allowing to use Ajax call to update any parts of
    subforms.
- Storage clients:
  - New interfaces and base classes for categories of storage clients:
    - RestClientInterface and RestClient (previously just "Rest") to implement
      any custom REST storage client. Note: the "base" class RestClient can be
      used "as is" (ie. not an abstract class) and is therefore not suffixed
      "Base".
    - FileClientInterface and FileClientBase to implement any file storage
      client.
    - QueryLanguageClientInterface and QueryLanguageClientBase to implement any
      query language or database storage client.
- Aggregator plugin: a new aggregator plugin has been added to allow aggregation
  of multiple storage clients (similar to xnttmulti plugin used to do for v2).
- REST client:
  - New endpoint options to support specific URLs to fetch a single entity or to
    count entities.
  - Support for any response structure using JSONPath to locate data in result
    sets.
  - Improved pager settings to support server limits and paging starting from 1.
  - Better support for authentication. Bearer token built-in support. API key
    support through query string. Support for custom authentication header.
  - Added support for custom headers. Ex.: "Accept: application/vnd.api+json".
  - Additional parameters can now be sent either as a query string or as POST
    data using the endpoint data encoding format. It is possible to mix POST and
    GET parameters using a question mark to specify query parameters in the
    endpoint URL.
  - Added an entity identifier placeholder "{id}" replaced in single entity
    endpoint URL and parameters.
- Files storage client is part of external entities module.
  - The files storage plugin ID changed from 'xnttfiles' to just 'files'.

### Upgrading v2 storage clients

- In the module's composer.json (if one) update requirements ("require") from
  `"drupal/external_entities": "^2.0@alpha"`
  (or similar) to
  `"drupal/external_entities": "^3.0@alpha"`
- In the storage class, update annotations from:
  ` * @ExternalEntityStorageClient(`
  to
  ` * @StorageClient(`
- Adjust ExternalEntityInterface and ExternalEntityTypeInterface namespaces.
- If the storage client implements the constructor (`__construct()`) and the
  `create()` methods, ResponseDecoderFactoryInterface has been removed and
  two other services are added: LoggerChannelFactoryInterface, and
  EntityFieldManagerInterface. Make the changes in both methods unless you use
  ResponseDecoderFactoryInterface. In that later case, you will need to do the
  changes in the __constructor() when calling parent class constructor and
  add LoggerChannelFactoryInterface, and EntityFieldManagerInterface to your
  __constructor() and create() methods. You can get a logger channel factory and
  an entity field manager from core services using:
  ```
  $container->get('logger.factory')
  $container->get('entity_field.manager')
  ```
  Don't forget to update code documentation and to add the import:
  ```
  use Drupal\Core\Logger\LoggerChannelFactoryInterface;
  use Drupal\Core\Entity\EntityFieldManagerInterface;
  ```
  For the explanations, ResponseDecoderFactoryInterface was only needed by REST
  clients and was removed from the StorageClientInterface. On the other side,
  EntityFieldManagerInterface was needed by the new features added to
  EntityFieldManagerInterface for query filtering. And the logger factory should
  be provided to all storage clients as they all would need to log things.
- To log messages, you can now use `$this->logger->notice(...)` or
  `$this->logger->warning(...)` or whatever log method you need.
- If the storage client extends the built-in REST client, you will have to take
  into account new configuration items added as well as form element changes.
  Also note the Rest storage client class has been renamed into "RestClient" for
  consistency (regarding other base clients).
- You should provide an implementation for `submitConfigurationForm()` and move
  some parts of `validateConfigurationForm()` into this new handler that should
  call `$this->setConfiguration($form_state->getValues());` in the end:
  ```
  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    ...
    $this->setConfiguration($form_state->getValues());
  }
  ```
- Rename query() and countQuery() into querySource() and countQuerySource()
  respectively. This should be sufficient unless you already tried to translate
  Drupal field names into mapped source field names in your code logic. In that
  case, you may need to do some more code adjustments.
- Implement transliterateDrupalFilters() method. To start, you can use the
  basic but unefficient implementation:
  ```
  public function transliterateDrupalFilters(array $parameters) :array {
    // @todo: To really implement for better performances.
    return ['source' => [], 'drupal' => $parameters];
  }
  ```
  When implementing, you may find helpfull to use
  $this->externalEntityType->getFieldMapper() and $this->entityFieldManager new
  class member as well as the getFieldDefinitions() and getFieldMappings() new
  methods. The storage client "knows" which operators are supported by its
  storage source (list of used operators: see StorageClientInterface::query()
  documentation and Drupal Query API documentation:
  https://www.drupal.org/docs/8/api/database-api/dynamic-queries/conditions and
  https://api.drupal.org/api/drupal/core%21lib%21Drupal%21Core%21Entity%21Query%21QueryInterface.php/function/QueryInterface%3A%3Acondition/10).
  Then, some operators may or may not be supported by some source fields. While
  the $parameters parameter containes filter on *Drupal* field names, it will be
  required to match those to source fields using the field mapper
  ($this->externalEntityType->getFieldMapper() and getFieldMappings() method).
  And to tell if a given source field can support a given operator, you may rely
  on the Drupal mapped field type given by the user to guess the source field
  type ($this->entityFieldManager and getFieldDefinitions() method).
- For derived REST clients:
  - You should get familiar with the new features and see if some may not change
    your approach.
  - Update buildConfigurationForm() to adjust elements that should be visible or
    hidden with pre-filled values. You may also adjust existing and new field
    properties.
  - If overriden, update getListQueryParameters() signature to:
    ```
    protected function getListQueryParameters(array $filters = []) {
    ```
    and replace `$parameters` by `$filters`.
    Also remove the use of getPagingQueryParameters() in that method as it is
    handled separatly now.
  - If overriden, update getPagingQueryParameters() logic according to the new
    behavior. Maybe this new implementation would not need to be overriden and
    you may just remove the override.
  - If overriden, update getSingleQueryParameters() signature to:
    ```
    protected function getSingleQueryParameters(int|string|null $id) :array {
    ```
  - If overriden, update getHttpHeaders() signature to:
    ```
    public function getHttpHeaders(
      string $request_type = 'list',
      string $endpoint = '',
      array $parameters = []
    ) :array {
    ```
- For derived File clients:
  - Now xnttfiles plugin is part of external entities. You need to remove the
    dependency in you .info.yml file:
    ```
    dependencies:
    - external_entities:external_entities
    - xnttfiles:xnttfiles
    ```
    becomes:
    ```
    dependencies:
    - external_entities:external_entities
    ```
    and your module composer.json needs to be updated as well to require only
    external_entities v3:
    ```
    "require": {
      "drupal/external_entities": "^2.0@beta",
      "drupal/xnttfiles": "^1.0@beta"
    }
    ```
    becomes:
    ```
    "require": {
      "drupal/external_entities": "^3.0@beta"
    }
    ```
  - Method ::getIdField() has been removed and can be replaced by base
    class method ::getSourceIdFieldName().
  - The internal raw field added to store file path has been renamed from
    '_xnttfiles_path' to '_xntt_path'.
  - The config structure changed a little (see
    FileClientBase::defaultConfiguration() and Files::defaultConfiguration()).
    You may need to implement a hook_update_N() in yourmodule.install file to
    manage settings upgrade. You will need to make sure your upgrade will occure
    after the external entity module corresponding one (93010). For example:
    ```
    /**
     * Implements hook_update_dependencies().
     */
    function yourmodule_update_dependencies() {
      $dependencies['yourmodule'][9101] = [
        'external_entities' => 93010,
      ];
      return $dependencies;
    }

    /**
     * Update field mapper configs to new structure.
     */
    function yourmodule_update_9101(&$sandbox) {
      $xntt_storage = \Drupal::entityTypeManager()
        ->getStorage('external_entity_type');
      $xntt_types = $xntt_storage
        ->getQuery()
        ->accessCheck(FALSE)
        ->execute();

      // Loop on external entity types.
      foreach ($xntt_types as $entity_type_id) {
        // Upgrade file storage client ids.
        /** @var \Drupal\Core\Config\Config $config */
        $config = \Drupal::configFactory()->getEditable(
          "external_entities.external_entity_type.{$entity_type_id}"
        );

        $scs = $config->get('storage_clients');
        $need_update = FALSE;
        foreach ($scs as &$sc) {
          if ($sc['id'] == 'yourmodule_pluginid') {
            $need_update = TRUE;
            $sc['config']['performances']['use_index'] =
              $sc['config']['use_index'];
            unset($sc['config']['use_index']);
            $sc['config']['performances']['index_file'] =
              $sc['config']['index_file'];
            unset($sc['config']['index_file']);
            $sc['config']['data_fields']['index_file'] =
              $sc['config']['field_list'];
            unset($sc['config']['field_list']);
            //... do some more stuff if needed.
          }
        }
        if ($need_update) {
          $config->set('storage_clients', $scs);
          $config->save();
        }
      }
    }
    ```

### Upgrading v2 field mappers

The field mapper approach changed a little: before, each field had a mapping
that could be either an expression or a constant (static) string value. To tell
which type of mapping was used, the field mapper system was using a prefix for
constants. It was not very convenient and was limiting the behavior to 2 cases:
constant and expression.

The new approach gets ride of the prefix and offers instead a drop down for each
field letting the user set if the field is mapped to a constant value or
something else (either an expression or a field name), which brings more
flexibility. Therefore, the previous field mapping value is not a mapping
string anymore but rather an array with 2 keys:
- 'mapping' containing the mapping string and
- 'type' containing the mapping type value as an integer corresponding to
  constants defined in FieldMapperInterface interface
  (src/FieldMapper/FieldMapperInterface.php) as MAPPING_* constants.
Therefore, the field mapper pluging interface must reflect that change and
either provide a way to select the mapping type (dropdown, radios), or provide
a pre-selected mapping type as a hidden value.

For more practical changes:
- In the module's composer.json (if one) update requirements ("require") from
  `"drupal/external_entities": "^2.0@alpha"`
  (or similar) to
  `"drupal/external_entities": "^3.0@alpha"`
- The base class of field mapper plugins is not
  `ConfigurableExpressionFieldMapperBase` anymore but `FieldMapperBase`.
  Replace:
  ```
  use Drupal\external_entities\FieldMapper\ConfigurableExpressionFieldMapperBase;
  ...
  ... extends ConfigurableExpressionFieldMapperBase ...
  ```
  By
  ```
  use Drupal\external_entities\FieldMapper\FieldMapperBase;
  use Drupal\external_entities\FieldMapper\FieldMapperInterface;
  ...
  ... extends FieldMapperBase ...
  ```
  Adding FieldMapperInterface is required to use field mapping type constants.
- Adjust ExternalEntityInterface and ExternalEntityTypeInterface namespaces.
- The default constructor has an addition parameter:
  LoggerChannelFactoryInterface $logger_factory
- Remove the implementation of:
  - ::getConstantMappingPrefix() (not used anymore)
  - ::extractIdFromRawData() (implemented in base class)
  - ::getFieldMappings() (if there are no special needs)
- Add type hinting to methods according to the FieldMapperInterface interface.
- Update ::buildConfigurationForm() according to the changes explained above
  about mapping type selection unless you rely on base implementation.
- ::extractEntityValuesFromRawData(), ::extractFieldValuesFromRawData(), and
  ::extractFieldPropertyValuesFromRawData() method signatures changed, are all
  public and are not using a FieldDefinitionInterface interface anymore.
  You will have to update the logic accordingly and take into account the
  mapping type now. Note: mapped constant values are handled elsewhere so you
  do not have to worry about them in ::extractFieldPropertyValuesFromRawData()
  method.
- Implement ::getPossibleFieldPropertyMappingTypes() that returns the mapping
  types allowed for the given field.
- You may implement ::getMappedFieldName() if some expressions of your field
  mapper can be used to directly map a raw field. For instance, a JSONPath
  expression like '$.somefield' is a direct mapping to the field 'somefield'.
- To log messages, you can now use `$this->logger->notice(...)` or
  `$this->logger->warning(...)` or whatever log method you need.
- Implement a hook_update_N() in yourmodule.install file to manage settings
  upgrade. You will need to make sure your upgrade will occure after the
  external entity module corresponding one (9306). For example:
  ```
  /**
   * Implements hook_update_dependencies().
   */
  function yourmodule_update_dependencies() {
    $dependencies['yourmodule'][9101] = [
      'external_entities' => 9306,
    ];
    return $dependencies;
  }

  /**
   * Update field mapper configs to new structure.
   */
  function yourmodule_update_9101(&$sandbox) {
    $xntt_storage = \Drupal::entityTypeManager()->getStorage('external_entity_type');
    $xntt_types = $xntt_storage
      ->getQuery()
      ->accessCheck(FALSE)
      ->execute();

    // Loop on external entity types.
    foreach ($xntt_types as $entity_type_id) {
      // Upgrade field mapper configs.
      /** @var \Drupal\Core\Config\Config $config */
      $config = \Drupal::configFactory()->getEditable(
        "external_entities.external_entity_type.{$entity_type_id}"
      );

      $fms = $config->get('field_mappers') ?? [];
      $need_upgrade = FALSE;
      foreach ($fms as &$fm) {
        if ('yourmodule_pluginid' == $fm['id']) {
          $need_upgrade = TRUE;
          foreach (
            $fm['config']['field_mappings']
            as $field_name => &$properties
          ) {
            foreach ($properties as $property => &$mapping) {
              // Check for old configs that need upgrade.
              if (is_string($mapping)) {
                $mapping_type = FieldMapperInterface::MAPPING_EXPRESSION;
                if ('+' == $mapping[0]) {
                  $mapping = substr($mapping, 1);
                  $mapping_type = FieldMapperInterface::MAPPING_CONSTANT;
                }
                elseif (<some checks on the $mapping content>) {
                  // You may rewrite $mapping.
                  // And set the mapping type.
                  $mapping_type = FieldMapperInterface::MAPPING_DIRECT;
                }
                //... Add other needed tests.
                $mapping = [
                  'type' => $mapping_type,
                  'mapping' => $mapping,
                ];
              }
            }
          }
        }
      }
      if ($need_upgrade) {
        $config->set('field_mappers', $fms);
        $config->save();
      }
    }
  }
  ```

## Maintainers

Current maintainers:
- Raf Philtjens - [rp7](https://www.drupal.org/u/rp7)
- Clemens Tolboom - [clemens.tolboom](https://www.drupal.org/u/clemenstolboom)
- Hanno Lans - [Hanno](https://www.drupal.org/u/hanno)
- Rodrigo Aguilera - [rodrigoaguilera](https://www.drupal.org/u/rodrigoaguilera)
- Jelle Sebreghts - [Jelle_S](https://www.drupal.org/u/jelle_s)
- Patrick van Efferen - [pefferen](https://www.drupal.org/u/pefferen)
- Peter Grond - [pgrond](https://www.drupal.org/u/pgrond)
- Valentin Guignon - [guignonv](https://www.drupal.org/u/guignonv)

This project has been sponsored by:
- Attiks (Peter Droogmans) is no nonsense company with a highly skilled
  technical and graphical staff. We handle both on- and offline communication,
  web development and print design are our middle names. We work
  internationally and are based in Antwerp, Belgium.
- Datascape (Hanno Lans) is involved in Drupal projects with an advanced
  information architecture, with accessibility, open content and open source in
  mind.
- Kunstmuseum The Hague has a leading collection of modern and contemporary art,
  fashion and decorative arts. It is also the international 
  home of Piet Mondrian,
  with no fewer than 300 works by the famous Dutch artist in its collection.
-The Alliance Bioversity - CIAT (guignonv), a CGIAR center.
