<?php

namespace Drupal\external_entities\DataAggregator;

use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\Plugin\PluginDependencyTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\external_entities\Entity\ExternalEntityTypeInterface;
use Drupal\external_entities\Plugin\PluginDebugTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for external entity data aggregators.
 */
abstract class DataAggregatorBase extends PluginBase implements DataAggregatorInterface {

  // Normally, we'd just need \Drupal\Core\Entity\DependencyTrait here for
  // plugins. However, in a few cases, plugins use plugins themselves, and then
  // the additional self::calculatePluginDependencies() method from this trait
  // is useful. Since PHP 5 complains when adding this trait along with its
  // "parent" trait to the same class, we just add it here in case a child class
  // does need it.
  use PluginDependencyTrait;
  use PluginDebugTrait;

  /**
   * The external entity type this storage client is configured for.
   *
   * @var \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   */
  protected $externalEntityType;

  /**
   * The field mapper manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $fieldMapperManager;

  /**
   * The external storage client manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $storageClientManager;

  /**
   * The logger channel factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerChannelFactory;

  /**
   * The storage client plugin logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannel
   */
  protected $logger;

  /**
   * Recursively merge 2 arrays.
   *
   * If $deep_merge is FALSE, only the first-level keys are taken into account.
   * In that case, sub-arrays of the first array may be overriden by values of
   * the second array (which may result in a non-array value).
   * Otherwise, when $deep_merge is TRUE, if the values of a given key is an
   * array either in the first or the second array, the resulting value will be
   * an array that merge the values from both sides, with overrides if needed.
   *
   * @param array $a
   *   First array that may be overriden.
   * @param array $b
   *   Second array that overrides existing values.
   * @param bool $deep_merge
   *   If TRUE, merge sub-arrays together.
   * @param bool $override_empty
   *   If TRUE, only first array empty values (WARNING: are considered empty
   *   NULL, '' and [] which differs from PHP empty() function) can be overriden
   *   by a second array non-empty value.
   * @param bool $preserve_integer_keys
   *   If TRUE, integer keys are preserved and values with the same index may be
   *   overriden according to the above rules.
   * @param bool $no_new_keys
   *   Only existing keys can be overriden. No new keys are added. If TRUE,
   *   parameter $preserve_integer_keys will be forced to TRUE.
   *
   * @return array
   *   The merged array.
   */
  public static function mergeArrays(
    array $a,
    array $b,
    bool $deep_merge = FALSE,
    bool $override_empty = FALSE,
    bool $preserve_integer_keys = FALSE,
    bool $no_new_keys = FALSE,
  ) :array {
    // Force $preserve_integer_keys if $no_new_keys is set.
    if ($no_new_keys) {
      $preserve_integer_keys = TRUE;
    }

    // Prepare result array.
    if ($preserve_integer_keys) {
      $result = $a;
    }
    else {
      $result = [];
      foreach ($a as $key => $value) {
        if (is_int($key)) {
          $result[] = $value;
        }
        else {
          $result[$key] = $value;
        }
      }
    }

    // Merge array b on result.
    if ($deep_merge) {
      // Merge sub-arrays.
      if ($override_empty) {
        foreach ($b as $key => $value) {
          // Check for new keys allowed.
          if ($no_new_keys && !array_key_exists($key, $result)) {
            continue;
          }

          if (is_int($key) && !$preserve_integer_keys) {
            // If we don't preserve int keys, we add the value.
            $result[] = $value;
          }
          elseif (((!isset($result[$key])) || ('' === $result[$key]) || ([] === $result[$key]))
            && ((NULL !== $value) && ('' !== $value) && ([] !== $value))
          ) {
            // Either the key is not int or it is but we preserve in keys, only
            // override empty values with non-empty ones.
            $result[$key] = $value;
          }
          elseif (((isset($result[$key])) && ('' !== $result[$key]) && ([] !== $result[$key]))
            && ((NULL !== $value) && ('' !== $value) && ([] !== $value))
          ) {
            if (is_array($result[$key])) {
              if (is_array($value)) {
                $result[$key] = static::mergeArrays(
                  $result[$key],
                  $value,
                  $deep_merge,
                  $override_empty,
                  $preserve_integer_keys,
                  $no_new_keys
                );
              }
              else {
                $result[$key] = static::mergeArrays(
                  $result[$key],
                  [$value],
                  $deep_merge,
                  $override_empty,
                  $preserve_integer_keys,
                  $no_new_keys
                );
              }
            }
            elseif (is_array($value)) {
              $result[$key] = static::mergeArrays(
                [$result[$key]],
                $value,
                $deep_merge,
                $override_empty,
                $preserve_integer_keys,
                $no_new_keys
              );
            }
          }
        }
      }
      else {
        // Full override.
        foreach ($b as $key => $value) {
          // Check for new keys allowed.
          if ($no_new_keys && !array_key_exists($key, $result)) {
            continue;
          }

          if (is_int($key) && !$preserve_integer_keys) {
            $result[] = $value;
          }
          elseif (isset($result[$key]) && is_array($result[$key])) {
            if (isset($value) && is_array($value)) {
              $result[$key] = static::mergeArrays(
                $result[$key],
                $value,
                $deep_merge,
                $override_empty,
                $preserve_integer_keys,
                $no_new_keys
              );
            }
            else {
              $result[$key] = static::mergeArrays(
                $result[$key],
                [$value],
                $deep_merge,
                $override_empty,
                $preserve_integer_keys,
                $no_new_keys
              );
            }
          }
          elseif (isset($result[$key]) && isset($value) && is_array($value)) {
            $result[$key] = static::mergeArrays(
              [$result[$key]],
              $value,
              $deep_merge,
              $override_empty,
              $preserve_integer_keys,
              $no_new_keys
            );
          }
          else {
            $result[$key] = $value;
          }
        }
      }
    }
    else {
      // Not merging sub-arrays.
      if ($override_empty) {
        foreach ($b as $key => $value) {
          // Check for new keys allowed.
          if ($no_new_keys && !array_key_exists($key, $result)) {
            continue;
          }

          if (is_int($key) && !$preserve_integer_keys) {
            // If we don't preserve int keys, we add the value.
            $result[] = $value;
          }
          elseif (((!isset($result[$key])) || ('' === $result[$key]) || ([] === $result[$key]))
            && ((NULL !== $value) && ('' !== $value) && ([] !== $value))
          ) {
            // Either the key is not int or it is but we preserve in keys, only
            // override empty values with non-empty ones.
            $result[$key] = $value;
          }
        }
      }
      else {
        // Full override.
        foreach ($b as $key => $value) {
          // Check for new keys allowed.
          if ($no_new_keys && !array_key_exists($key, $result)) {
            continue;
          }

          if (is_int($key) && !$preserve_integer_keys) {
            $result[] = $value;
          }
          else {
            $result[$key] = $value;
          }
        }
      }
    }
    return $result;
  }

  /**
   * Constructs a DataAggregatorBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger channel factory.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $field_mapper_manager
   *   The field mapper manager.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $storage_client_manager
   *   The storage client manager.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    $plugin_definition,
    TranslationInterface $string_translation,
    LoggerChannelFactoryInterface $logger_factory,
    PluginManagerInterface $field_mapper_manager,
    PluginManagerInterface $storage_client_manager,
  ) {
    $configuration += $this->defaultConfiguration();
    if (!empty($configuration[ExternalEntityTypeInterface::XNTT_TYPE_PROP])
        && $configuration[ExternalEntityTypeInterface::XNTT_TYPE_PROP] instanceof ExternalEntityTypeInterface
    ) {
      $this->externalEntityType = $configuration[ExternalEntityTypeInterface::XNTT_TYPE_PROP];
      unset($configuration[ExternalEntityTypeInterface::XNTT_TYPE_PROP]);
    }
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->setStringTranslation($string_translation);
    $this->loggerChannelFactory = $logger_factory;
    $this->logger = $this->loggerChannelFactory->get('xntt_data_aggregator_' . $plugin_id);
    $this->fieldMapperManager = $field_mapper_manager;
    $this->storageClientManager = $storage_client_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('string_translation'),
      $container->get('logger.factory'),
      $container->get('plugin.manager.external_entities.field_mapper'),
      $container->get('plugin.manager.external_entities.storage_client')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getLabel() :string {
    $plugin_definition = $this->getPluginDefinition();
    return $plugin_definition['label'];
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() :string {
    $plugin_definition = $this->getPluginDefinition();
    return $plugin_definition['description'] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = static::mergeArrays(
      $this->defaultConfiguration(),
      $configuration,
      TRUE,
      TRUE,
      TRUE,
      FALSE,
    );
    if (array_key_exists(ExternalEntityTypeInterface::XNTT_TYPE_PROP, $this->configuration)) {
      $this->externalEntityType = $this->configuration[ExternalEntityTypeInterface::XNTT_TYPE_PROP];
      unset($this->configuration[ExternalEntityTypeInterface::XNTT_TYPE_PROP]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'global_config' => [],
      'storage_client_config' => [
        [],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function buildGlobalAggregationForm(array $form, FormStateInterface $form_state) {
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function buildStorageClientAggregationForm(array $form, FormStateInterface $form_state, int $client_index) {
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateGlobalAggregationForm(array &$form, FormStateInterface $form_state) {
  }

  /**
   * {@inheritdoc}
   */
  public function validateStorageClientAggregationForm(array &$form, FormStateInterface $form_state, int $client_index) {
  }

  /**
   * {@inheritdoc}
   */
  public function submitGlobalAggregationForm(array &$form, FormStateInterface $form_state) {
    $config = $this->getConfiguration();
    $config['global_config'] = $form_state->getValues();
    $this->setConfiguration($config);
  }

  /**
   * {@inheritdoc}
   */
  public function submitStorageClientAggregationForm(array &$form, FormStateInterface $form_state, int $client_index) {
    $config = $this->getConfiguration();
    $config['storage_client_config'][$client_index] = $form_state->getValues();
    $this->setConfiguration($config);
  }

}
