<?php

namespace Drupal\external_entities\Entity;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\external_entities\DataAggregator\DataAggregatorInterface;
use Drupal\external_entities\DataAggregator\DataAggregatorManager;
use Drupal\external_entities\FieldMapper\FieldMapperInterface;
use Drupal\external_entities\FieldMapper\FieldMapperManager;
use Drupal\external_entities\StorageClient\StorageClientInterface;
use Drupal\external_entities\StorageClient\StorageClientManager;
use Psr\Log\LoggerInterface;

/**
 * Defines the external_entity_type entity.
 *
 * @ConfigEntityType(
 *   id = "external_entity_type",
 *   label = @Translation("External entity type"),
 *   handlers = {
 *     "list_builder" = "Drupal\external_entities\ExternalEntityTypeListBuilder",
 *     "form" = {
 *       "add" = "Drupal\external_entities\Form\ExternalEntityTypeForm",
 *       "edit" = "Drupal\external_entities\Form\ExternalEntityTypeForm",
 *       "delete" = "Drupal\external_entities\Form\ExternalEntityTypeDeleteForm",
 *     }
 *   },
 *   config_prefix = "external_entity_type",
 *   admin_permission = "administer external entity types",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label",
 *   },
 *   links = {
 *     "edit-form" = "/admin/structure/external-entity-types/{external_entity_type}",
 *     "delete-form" = "/admin/structure/external-entity-types/{external_entity_type}/delete",
 *   },
 *   config_export = {
 *     "id",
 *     "label",
 *     "label_plural",
 *     "description",
 *     "content_class",
 *     "generate_aliases",
 *     "read_only",
 *     "debug_level",
 *     "field_mappers",
 *     "storage_clients",
 *     "data_aggregator",
 *     "persistent_cache_max_age",
 *     "annotation_entity_type_id",
 *     "annotation_bundle_id",
 *     "annotation_field_name",
 *     "inherits_annotation_fields"
 *  }
 * )
 */
class ExternalEntityType extends ConfigEntityBase implements ExternalEntityTypeInterface {

  /**
   * Indicates that entities of this external entity type should not be cached.
   */
  const CACHE_DISABLED = 0;

  /**
   * Default field mapper plugin id.
   */
  const DEFAULT_FIELD_MAPPER = 'simple';

  /**
   * Default data aggregator plugin id.
   */
  const DEFAULT_DATA_AGGREGATOR = 'group';

  /**
   * Default storage client plugin id.
   */
  const DEFAULT_STORAGE_CLIENT = 'rest';

  /**
   * The external entity type ID.
   *
   * @var string
   */
  protected $id;

  /**
   * The human-readable name of the external entity type.
   *
   * @var string
   */
  protected $label;

  /**
   * The plural human-readable name of the external entity type.
   *
   * @var string
   */
  protected $label_plural;

  /**
   * The external entity type description.
   *
   * @var string
   */
  protected $description;

  /**
   * The external entity class used for content instances.
   *
   * @var string
   */
  protected $content_class = ExternalEntity::class;

  /**
   * Whether or not entity types of this external entity type are read only.
   *
   * @var bool
   */
  protected $read_only;

  /**
   * Debug level of current instance.
   *
   * @var int
   */
  protected $debug_level;

  /**
   * Whether or not to auto-generate aliases for this external entity type.
   *
   * @var bool
   */
  protected $generate_aliases;

  /**
   * Array of field mapper plugin settings.
   *
   * @var array
   */
  protected $field_mappers = [];

  /**
   * The field mapper plugin instance.
   *
   * @var \Drupal\external_entities\FieldMapper\FieldMapperInterface
   */
  protected $fieldMapperPlugins = [];

  /**
   * Array of storage client plugin settings.
   *
   * @var array
   */
  protected $storage_clients = [];

  /**
   * Array of storage client plugin instances.
   *
   * @var \Drupal\external_entities\StorageClient\StorageClientInterface[]
   */
  protected $storageClientPlugins = [];

  /**
   * Data aggregator plugin settings.
   *
   * @var array
   */
  protected $data_aggregator = [];

  /**
   * Data aggregator plugin instance.
   *
   * @var \Drupal\external_entities\DataAggregator\DataAggregatorInterface
   */
  protected $dataAggregator;

  /**
   * Max age entities of this external entity type may be persistently cached.
   *
   * @var int
   */
  protected $persistent_cache_max_age = self::CACHE_DISABLED;

  /**
   * The annotations entity type id.
   *
   * @var string
   */
  protected $annotation_entity_type_id;

  /**
   * The annotations bundle id.
   *
   * @var string
   */
  protected $annotation_bundle_id;

  /**
   * The field this external entity is referenced from by the annotation entity.
   *
   * @var string
   */
  protected $annotation_field_name;

  /**
   * Local cache for the annotation field.
   *
   * @var array
   *
   * @see ExternalEntityType::getAnnotationField()
   */
  protected $annotationField;

  /**
   * Indicates if the external entity inherits the annotation entity fields.
   *
   * @var bool
   */
  protected $inherits_annotation_fields = FALSE;

  /**
   * {@inheritdoc}
   */
  public function getLabel() :string {
    return $this->label ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function getPluralLabel() :string {
    return $this->label_plural ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() :string {
    return $this->description ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function isReadOnly() :bool {
    return (bool) $this->read_only;
  }

  /**
   * {@inheritdoc}
   */
  public function getDebugLevel() :int {
    return $this->debug_level ?? 0;
  }

  /**
   * {@inheritdoc}
   */
  public function setDebugLevel(int $debug_level = 1) {
    $this->debug_level = $debug_level;
  }

  /**
   * {@inheritdoc}
   */
  public function automaticallyGenerateAliases() :bool {
    return (bool) $this->generate_aliases;
  }

  /**
   * {@inheritdoc}
   */
  public function hasValidFieldMapper(int $mapper_number = 0) :bool {
    if ((0 > $mapper_number)
        || ($mapper_number >= $this->getFieldMapperCount())
    ) {
      return FALSE;
    }
    $field_mapper_plugin_definition =
      \Drupal::service('plugin.manager.external_entities.field_mapper')
        ->getDefinition($this->getFieldMapperId($mapper_number), FALSE);
    return !empty($field_mapper_plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMapperCount() :int {
    return count($this->field_mappers);
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMapperId(int $mapper_number = 0) :string {
    return $this->field_mappers[$mapper_number]['id'] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function setFieldMapperId(
    string $field_mapper_id,
    int $mapper_number = 0
  ) :self {
    if (!$mapper_number
        || ((0 <= $mapper_number)
            && ($mapper_number < $this->getFieldMapperCount()))
    ) {
      $this->field_mappers[$mapper_number] ??= ['config' => []];
      $this->field_mappers[$mapper_number]['id'] = $field_mapper_id;
      $this->fieldMapperPlugins[$mapper_number] = NULL;
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function addFieldMapperId(
    string $field_mapper_id
  ) :self {
    $this->field_mappers[] = [
      'id' => $field_mapper_id,
      'config' => [],
    ];
    $this->fieldMapperPlugins[] = NULL;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function removeFieldMapperId(
    int $mapper_number = 0
  ) :self {
    if ((0 <= $mapper_number)
        && ($mapper_number < $this->getFieldMapperCount())
    ) {
      array_splice(
        $this->field_mappers,
        $mapper_number,
        1
      );
      array_splice(
        $this->fieldMapperPlugins,
        $mapper_number,
        1
      );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMapper(
    int $mapper_number = 0
  ) :FieldMapperInterface {
    if ((0 > $mapper_number)
        || ((0 < $mapper_number)
          && ($mapper_number >= $this->getFieldMapperCount()))
    ) {
      throw new PluginException(
        sprintf('Invalid field mapper plugin number %d.', $mapper_number)
      );
    }
    if (empty($this->fieldMapperPlugins[$mapper_number])) {
      $field_mapper_plugin_manager =
        \Drupal::service('plugin.manager.external_entities.field_mapper');
      assert($field_mapper_plugin_manager instanceof FieldMapperManager);

      $config = $this->getFieldMapperConfig($mapper_number);

      // Allow the mapper to call back into the entity type (e.g., to fetch
      // additional data like field lists from the remote service).
      $config[ExternalEntityTypeInterface::XNTT_TYPE_PROP] = $this;

      $this->fieldMapperPlugins[$mapper_number] =
        $field_mapper_plugin_manager->createInstance(
          $this->getFieldMapperId($mapper_number),
          $config
        );
      // Adjust field mapper debug level to current one only if set.
      if ($debug_level = $this->getDebugLevel()) {
        $this->fieldMapperPlugins[$mapper_number]->setDebugLevel($debug_level);
      }
    }

    return $this->fieldMapperPlugins[$mapper_number];
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMappers() :array {
    if (empty($this->fieldMapperPlugins) && $this->getFieldMapperCount()) {
      $field_mapper_plugin_manager =
        \Drupal::service('plugin.manager.external_entities.field_mapper');
      assert($field_mapper_plugin_manager instanceof FieldMapperManager);

      for ($mapper_number = 0; $mapper_number < $this->getFieldMapperCount(); ++$mapper_number) {
        $config = $this->getFieldMapperConfig($mapper_number);

        // Allow the mapper to call back into the entity type (e.g., to fetch
        // additional data like field lists from the remote service).
        $config[ExternalEntityTypeInterface::XNTT_TYPE_PROP] = $this;

        $this->fieldMapperPlugins[$mapper_number] =
          $field_mapper_plugin_manager->createInstance(
            $this->getFieldMapperId($mapper_number),
            $config
          );
        // Adjust field mapper debug level to current one only if set.
        if ($debug_level = $this->getDebugLevel()) {
          $this->fieldMapperPlugins[$mapper_number]->setDebugLevel($debug_level);
        }
      }
    }

    return $this->fieldMapperPlugins;
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMapperConfig(int $mapper_number = 0) :array {
    return $this->field_mappers[$mapper_number]['config'] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function setFieldMapperConfig(
    array $field_mapper_config,
    int $mapper_number = 0
  ) :self {
    if ((0 <= $mapper_number)
        && ($mapper_number < $this->getFieldMapperCount())
    ) {
      // Update local config.
      $this->field_mappers[$mapper_number]['config'] = $field_mapper_config;
      // Update plugin config.
      $this
        ->getFieldMapper($mapper_number)
        ->setConfiguration(
          $field_mapper_config
          + [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $this]
        );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMapperNotes(int $mapper_number = 0) :string {
    return $this->field_mappers[$mapper_number]['notes'] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function setFieldMapperNotes(
    string $field_mapper_notes,
    int $mapper_number = 0
  ) :self {
    if ((0 <= $mapper_number)
        && ($mapper_number < $this->getFieldMapperCount())
        && (!empty($this->field_mappers[$mapper_number]))
    ) {
      $this->field_mappers[$mapper_number]['notes'] = $field_mapper_notes;
    }
    elseif ($mapper_number) {
      $this->logger($this->id())->warning(
        'Trying to set notes for a field mapper out of current set (@mapper_number/@total).',
        [
          '@mapper_number' => $mapper_number,
          '@total' => $this->getFieldMapperCount(),
        ]
      );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function hasValidStorageClient(int $client_number = 0) :bool {
    if ((0 > $client_number)
        || ($client_number >= $this->getStorageClientCount())
    ) {
      return FALSE;
    }
    $storage_client_plugin_definition =
      \Drupal::service('plugin.manager.external_entities.storage_client')
        ->getDefinition($this->getStorageClientId($client_number), FALSE);
    return !empty($storage_client_plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public function getStorageClientCount() :int {
    return count($this->storage_clients);
  }

  /**
   * {@inheritdoc}
   */
  public function getStorageClientId(int $client_number = 0) :string {
    return $this->storage_clients[$client_number]['id'] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function setStorageClientId(
    string $storage_client_id,
    int $client_number = 0
  ) :self {
    if (!$client_number
        || ((0 <= $client_number)
            && ($client_number < $this->getStorageClientCount()))
    ) {
      $this->storage_clients[$client_number] ??= ['config' => []];
      $this->storage_clients[$client_number]['id'] = $storage_client_id;
      $this->storageClientPlugins[$client_number] = NULL;
    }
    elseif ($client_number) {
      $this->logger($this->id())->warning(
        'Trying to set a storage client out of current set (@client_number/@total).',
        [
          '@client_number' => $client_number,
          '@total' => $this->getStorageClientCount(),
        ]
      );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function addStorageClientId(
    string $storage_client_id
  ) :self {
    $this->storage_clients[] = [
      'id' => $storage_client_id,
      'config' => [],
    ];
    $this->storageClientPlugins[] = NULL;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function removeStorageClientId(
    int $client_number = 0
  ) :self {
    if ((0 <= $client_number)
        && ($client_number < $this->getStorageClientCount())
    ) {
      array_splice(
        $this->storage_clients,
        $client_number,
        1
      );
      array_splice(
        $this->storageClientPlugins,
        $client_number,
        1
      );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getStorageClient(int $client_number = 0)
  :StorageClientInterface {
    if ((0 > $client_number)
        || ($client_number >= $this->getStorageClientCount())
    ) {
      throw new PluginException(
        sprintf('Invalid storage client plugin number %d.', $client_number)
      );
    }
    if (empty($this->storageClientPlugins[$client_number])) {
      $storage_client_plugin_manager =
        \Drupal::service('plugin.manager.external_entities.storage_client');
      assert($storage_client_plugin_manager instanceof StorageClientManager);

      $config = $this->getStorageClientConfig($client_number);

      // Allow the storage client to call back into the entity type (e.g., to
      // fetch additional data or settings).
      $config[ExternalEntityTypeInterface::XNTT_TYPE_PROP] = $this;

      $this->storageClientPlugins[$client_number] =
        $storage_client_plugin_manager->createInstance(
          $this->getStorageClientId($client_number),
          $config
        );
      // Adjust storage client debug level to current one only if set.
      if ($debug_level = $this->getDebugLevel()) {
        $this->storageClientPlugins[$client_number]->setDebugLevel($debug_level);
      }
    }

    return $this->storageClientPlugins[$client_number];
  }

  /**
   * {@inheritdoc}
   */
  public function getStorageClients() :array {
    if (empty($this->storageClientPlugins) && $this->getStorageClientCount()) {
      $storage_client_plugin_manager =
        \Drupal::service('plugin.manager.external_entities.storage_client');
      assert($storage_client_plugin_manager instanceof StorageClientManager);

      for ($client_number = 0; $client_number < $this->getStorageClientCount(); ++$client_number) {
        $config = $this->getStorageClientConfig($client_number);

        // Allow the storage client to call back into the entity type (e.g., to
        // fetch additional data or settings).
        $config[ExternalEntityTypeInterface::XNTT_TYPE_PROP] = $this;

        $this->storageClientPlugins[$client_number] =
          $storage_client_plugin_manager->createInstance(
            $this->getStorageClientId($client_number),
            $config
          );
        // Adjust storage client debug level to current one only if set.
        if ($debug_level = $this->getDebugLevel()) {
          $this->storageClientPlugins[$client_number]->setDebugLevel($debug_level);
        }
      }
    }

    return $this->storageClientPlugins;
  }

  /**
   * {@inheritdoc}
   */
  public function getStorageClientConfig(int $client_number = 0) :array {
    return $this->storage_clients[$client_number]['config'] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function setStorageClientConfig(
    array $storage_client_config,
    int $client_number = 0
  ) :self {
    if ((0 <= $client_number)
        && ($client_number < $this->getStorageClientCount())
    ) {
      // Update local config.
      $this->storage_clients[$client_number]['config'] = $storage_client_config;
      // Update plugin config.
      $this
        ->getStorageClient($client_number)
        ->setConfiguration(
          $storage_client_config
          + [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $this]
        );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getStorageClientNotes(int $client_number = 0) :string {
    return $this->storage_clients[$client_number]['notes'] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function setStorageClientNotes(
    string $storage_client_notes,
    int $client_number = 0
  ) :self {
    if ((0 <= $client_number)
        && ($client_number < $this->getStorageClientCount())
        && (!empty($this->storage_clients[$client_number]))
    ) {
      $this->storage_clients[$client_number]['notes'] = $storage_client_notes;
    }
    elseif ($client_number) {
      $this->logger($this->id())->warning(
        'Trying to set notes for a storage client out of current set (@client_number/@total).',
        [
          '@client_number' => $client_number,
          '@total' => $this->getStorageClientCount(),
        ]
      );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDataAggregatorId() :string {
    return $this->data_aggregator['id'] ?? static::DEFAULT_DATA_AGGREGATOR;
  }

  /**
   * {@inheritdoc}
   */
  public function setDataAggregatorId(
    string $data_aggregator_id,
  ) :self {
    // @todo Update config if id changes.
    $this->data_aggregator['id'] = $data_aggregator_id;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDataAggregator() :DataAggregatorInterface {

    if (empty($this->dataAggregator)) {
      $data_aggregator_plugin_manager =
        \Drupal::service('plugin.manager.external_entities.data_aggregator');
      assert($data_aggregator_plugin_manager instanceof DataAggregatorManager);

      $config = $this->getDataAggregatorConfig();

      // Allow the aggregator to call back into the entity type to get storage
      // clients settings.
      $config[ExternalEntityTypeInterface::XNTT_TYPE_PROP] = $this;

      $aggregator_id = $this->getDataAggregatorId();
      // Additional check if not default aggregator.
      if (static::DEFAULT_DATA_AGGREGATOR != $aggregator_id) {
        $plugin_definition =
          $data_aggregator_plugin_manager->getDefinition($aggregator_id, FALSE);
        if (empty($plugin_definition)) {
          $aggregator_id = static::DEFAULT_DATA_AGGREGATOR;
          $this->setDataAggregatorId($aggregator_id);
        }
      }
      $this->dataAggregator = $data_aggregator_plugin_manager->createInstance(
        $aggregator_id,
        $config
      );

      // Adjust field mapper debug level to current one only if set.
      if ($debug_level = $this->getDebugLevel()) {
        $this->dataAggregator->setDebugLevel($debug_level);
      }
    }
    return $this->dataAggregator;
  }

  /**
   * {@inheritdoc}
   */
  public function getDataAggregatorConfig() :array {
    return $this->data_aggregator['config'] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function setDataAggregatorConfig(
    array $data_aggregator_config
  ) :self {
    $this->data_aggregator['config'] = $data_aggregator_config;
    $this->getDataAggregator()->setConfiguration(
      $data_aggregator_config
      + [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $this]
    );
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getStorageClientAggregationConfig(
    int $client_number
  ) :array {
    return $this->data_aggregator['config']['storage_client_config'][$client_number]
      ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function setStorageClientAggregationConfig(
    array $storage_client_aggr_config,
    int $client_number
  ) :self {
    if ((0 <= $client_number)
        && ($client_number < $this->getStorageClientCount())
    ) {
      // Update local config.
      $this
        ->data_aggregator['config']['storage_client_config'][$client_number] =
        $storage_client_aggr_config;
      // Update plugin config.
      $this->getDataAggregator()->setConfiguration(
        $this->data_aggregator['config']
        + [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $this]
      );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getPersistentCacheMaxAge() :int {
    return $this->persistent_cache_max_age;
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(EntityStorageInterface $storage, $update = TRUE) {
    parent::postSave($storage, $update);

    // Clear the entity type definitions cache so changes flow through to the
    // related entity types.
    $this->entityTypeManager()->clearCachedDefinitions();

    // Clear the router cache to prevent RouteNotFoundException errors caused
    // by the Field UI module.
    \Drupal::service('router.builder')->rebuild();

    // Rebuild local actions so that the 'Add field' action on the 'Manage
    // fields' tab appears.
    \Drupal::service('plugin.manager.menu.local_action')->clearCachedDefinitions();

    // Clear the static and persistent cache.
    $storage->resetCache();
    if ($this->entityTypeManager()->hasDefinition($this->id())) {
      $this
        ->entityTypeManager()
        ->getStorage($this->id())
        ->resetCache();
    }

    $edit_link = $this->toLink(t('Edit entity type'), 'edit-form')->toString();
    if ($update) {
      $this->logger($this->id())->notice(
        'Entity type %label has been updated.',
        ['%label' => $this->label(), 'link' => $edit_link]
      );
    }
    else {
      // Notify storage to create the database schema.
      $entity_type = $this->entityTypeManager()->getDefinition($this->id());
      \Drupal::service('entity_type.listener')
        ->onEntityTypeCreate($entity_type);

      $this->logger($this->id())->notice(
        'Entity type %label has been added.',
        ['%label' => $this->label(), 'link' => $edit_link]
      );
    }

  }

  /**
   * {@inheritdoc}
   */
  public static function postDelete(EntityStorageInterface $storage, array $entities) {
    parent::postDelete($storage, $entities);
    \Drupal::service('entity_type.manager')->clearCachedDefinitions();
  }

  /**
   * {@inheritdoc}
   */
  public function getDerivedEntityTypeId() :string {
    return $this->id();
  }

  /**
   * {@inheritdoc}
   */
  public function getDerivedEntityType() :ContentEntityTypeInterface|null {
    return $this->entityTypeManager()->getDefinition($this->getDerivedEntityTypeId(), FALSE);
  }

  /**
   * {@inheritdoc}
   */
  public function isAnnotatable() :bool {
    return $this->getAnnotationEntityTypeId()
      && $this->getAnnotationBundleId()
      && $this->getAnnotationFieldName();
  }

  /**
   * {@inheritdoc}
   */
  public function getAnnotationEntityTypeId() :string|null {
    return $this->annotation_entity_type_id;
  }

  /**
   * {@inheritdoc}
   */
  public function getAnnotationBundleId() :string|null {
    return $this->annotation_bundle_id ?: $this->getAnnotationEntityTypeId();
  }

  /**
   * {@inheritdoc}
   */
  public function getAnnotationFieldName() :string|null {
    return $this->annotation_field_name;
  }

  /**
   * Returns the entity field manager.
   *
   * @return \Drupal\Core\Entity\EntityFieldManagerInterface
   *   The entity field manager.
   */
  protected function entityFieldManager() :EntityFieldManagerInterface {
    return \Drupal::service('entity_field.manager');
  }

  /**
   * {@inheritdoc}
   */
  public function getAnnotationField() :FieldDefinitionInterface|null {
    if (!isset($this->annotationField) && $this->isAnnotatable()) {
      $field_definitions = $this->entityFieldManager()->getFieldDefinitions($this->getAnnotationEntityTypeId(), $this->getAnnotationBundleId());
      $annotation_field_name = $this->getAnnotationFieldName();
      if (!empty($field_definitions[$annotation_field_name])) {
        $this->annotationField = $field_definitions[$annotation_field_name];
      }
    }

    return $this->annotationField;
  }

  /**
   * {@inheritdoc}
   */
  public function inheritsAnnotationFields() :bool {
    return (bool) $this->inherits_annotation_fields;
  }

  /**
   * {@inheritdoc}
   */
  public function getBasePath() :string {
    return str_replace('_', '-', strtolower($this->id));
  }

  /**
   * Gets the logger for a specific channel.
   *
   * @param string $channel
   *   The name of the channel.
   *
   * @return \Psr\Log\LoggerInterface
   *   The logger for this channel.
   */
  protected function logger($channel) :LoggerInterface {
    return \Drupal::getContainer()->get('logger.factory')->get($channel);
  }

  /**
   * {@inheritdoc}
   */
  public function __sleep() {
    // Prevent some properties from being serialized.
    return array_diff(parent::__sleep(), [
      'fieldMapperPlugins',
      'storageClientPlugins',
      'annotationField',
    ]);
  }

}
