<?php

namespace Drupal\external_entities\Entity;

use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\external_entities\DataAggregator\DataAggregatorInterface;
use Drupal\external_entities\FieldMapper\FieldMapperInterface;
use Drupal\external_entities\StorageClient\StorageClientInterface;

/**
 * Provides an interface defining an external entity type entity.
 */
interface ExternalEntityTypeInterface extends ConfigEntityInterface {

  /**
   * Internal property name used to share external entity type with plugins.
   */
  const XNTT_TYPE_PROP = '_external_entity_type';

  /**
   * Gets the human-readable name of the entity type.
   *
   * This label should be used to present a human-readable name of the
   * entity type.
   *
   * @return string
   *   The human-readable name of the entity type.
   */
  public function getLabel() :string;

  /**
   * Gets the plural human-readable name of the entity type.
   *
   * This label should be used to present a plural human-readable name of the
   * entity type.
   *
   * @return string
   *   The plural human-readable name of the entity type.
   */
  public function getPluralLabel() :string;

  /**
   * Gets the description.
   *
   * @return string
   *   The external entity types description, or an empty string.
   */
  public function getDescription() :string;

  /**
   * Returns if entities of this external entity are read only.
   *
   * @return bool
   *   TRUE if the entities are read only, FALSE otherwise.
   */
  public function isReadOnly() :bool;

  /**
   * Returns debug level for current instance.
   *
   * @return int
   *   The debug level. 0 means debugging is off.
   */
  public function getDebugLevel() :int;

  /**
   * Sets debug level for current instance.
   *
   * @param int $debug_level
   *   Sets the debug level. A non-0 value means debugging is enabled while 0
   *   disables  debugging. Default to 1 (enabled).
   */
  public function setDebugLevel(int $debug_level = 1);

  /**
   * Returns if we need to automatically generate aliases for this entity.
   *
   * @return bool
   *   TRUE if we need to generate aliases, FALSE otherwise.
   */
  public function automaticallyGenerateAliases() :bool;

  /**
   * Determines whether the field mapper is valid.
   *
   * @param int $mapper_number
   *   The number of the field mapper. Default to 0.
   *
   * @return bool
   *   TRUE if the field mapper is valid, FALSE otherwise.
   */
  public function hasValidFieldMapper(int $mapper_number = 0) :bool;

  /**
   * Returns the number of field mappers used by this type.
   *
   * @return int
   *   Number of field mappers. 0 if none.
   */
  public function getFieldMapperCount() :int;

  /**
   * Retrieves the plugin ID of the given field mapper number for this type.
   *
   * @param int $mapper_number
   *   The number of the field mapper. Default to 0.
   *
   * @return string
   *   The plugin ID of the field mapper. If the requested mapper number is
   *   not set, returns an empty string as default value.
   */
  public function getFieldMapperId(int $mapper_number = 0) :string;

  /**
   * Sets the plugin ID of the given field mapper number for this type.
   *
   * If the mapper number is not 0 and is out of range, nothing is set.
   *
   * @param string $field_mapper_id
   *   The plugin ID of the field mapper.
   * @param int $mapper_number
   *   The number of the field mapper. Default to 0.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function setFieldMapperId(
    string $field_mapper_id,
    int $mapper_number = 0
  ) :self;

  /**
   * Adds a field mapper plugin ID to this type.
   *
   * @param string $field_mapper_id
   *   The plugin ID of the field mapper.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function addFieldMapperId(
    string $field_mapper_id
  ) :self;

  /**
   * Removes the plugin ID of the given field mapper number for this type.
   *
   * It removes the mapper identifier from the list and shifts the array of
   * field mapper identifiers and configs if needed to avoid gaps. It does
   * nothing if the mapper number is out of range.
   *
   * @param int $mapper_number
   *   The number of the field mapper. Default to 0.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function removeFieldMapperId(
    int $mapper_number = 0
  ) :self;

  /**
   * Retrieves the requested field mapper.
   *
   * @param int $mapper_number
   *   The number of the field mapper. Default to 0.
   *
   * @return \Drupal\external_entities\FieldMapper\FieldMapperInterface
   *   The requested field mapper plugin.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   Thrown if the field mapper plugin could not be retrieved or if the
   *   given mapper number is out of range (therefore the mapper can't be
   *   found).
   */
  public function getFieldMapper(int $mapper_number = 0)
  :FieldMapperInterface;

  /**
   * Retrieves all the field mappers.
   *
   * @return \Drupal\external_entities\FieldMapper\FieldMapperInterface[]
   *   The field mapper plugins.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   Thrown if a field mapper plugin could not be retrieved.
   */
  public function getFieldMappers() :array;

  /**
   * Retrieves the configuration of this field mapper plugin.
   *
   * @param int $mapper_number
   *   The number of the field mapper. Default to 0.
   *
   * @return array
   *   An associative array with the field mapper configuration. If the
   *   requested mapper number has no config set, returns an empty array as
   *   default value.
   */
  public function getFieldMapperConfig(int $mapper_number = 0) :array;

  /**
   * Sets the configuration of the given field mapper plugin.
   *
   * If the mapper number is out of range, nothing is set.
   *
   * @param array $field_mapper_config
   *   The new configuration for the field mappert.
   * @param int $mapper_number
   *   The number of the field mapper. Default to 0.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function setFieldMapperConfig(
    array $field_mapper_config,
    int $mapper_number = 0
  ) :self;

  /**
   * Retrieves the administrative notes for the given field mapper number.
   *
   * @param int $mapper_number
   *   The number of the field mapper. Default to 0.
   *
   * @return string
   *   The administratives notes of the given field mapper. If the requested
   *   mapper number is not set, returns an empty string as default value.
   */
  public function getFieldMapperNotes(int $mapper_number = 0) :string;

  /**
   * Sets the administrative notes for the given field mapper number.
   *
   * If the mapper number is not 0 and is out of range, nothing is set.
   *
   * @param string $field_mapper_notes
   *   The administrative notes for the given field mapper.
   * @param int $mapper_number
   *   The number of the field mapper. Default to 0.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function setFieldMapperNotes(
    string $field_mapper_notes,
    int $mapper_number = 0
  ) :self;

  /**
   * Determines whether the storage client is valid.
   *
   * @param int $client_number
   *   The number of the client storage. Default to 0.
   *
   * @return bool
   *   TRUE if the storage client is valid, FALSE otherwise.
   */
  public function hasValidStorageClient(int $client_number = 0) :bool;

  /**
   * Returns the number of storage clients used by this type.
   *
   * @return int
   *   Number of storage clients. 0 if none.
   */
  public function getStorageClientCount() :int;

  /**
   * Retrieves the plugin ID of the given storage client number for this type.
   *
   * @param int $client_number
   *   The number of the client storage. Default to 0.
   *
   * @return string
   *   The plugin ID of the storage client. If the requested client number is
   *   not set, returns an empty string as default value.
   */
  public function getStorageClientId(int $client_number = 0) :string;

  /**
   * Sets the plugin ID of the given storage client number for this type.
   *
   * If the client number is not 0 and is out of range, nothing is set.
   *
   * @param string $storage_client_id
   *   The plugin ID of the storage client.
   * @param int $client_number
   *   The number of the client storage. Default to 0.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function setStorageClientId(
    string $storage_client_id,
    int $client_number = 0
  ) :self;

  /**
   * Adds a storage client plugin ID to this type.
   *
   * @param string $storage_client_id
   *   The plugin ID of the storage client.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function addStorageClientId(
    string $storage_client_id
  ) :self;

  /**
   * Removes the plugin ID of the given storage client number for this type.
   *
   * It removes the client identifier from the list and shifts the array of
   * storage client identifiers and configs if needed to avoid gaps. It does
   * nothing if the client number is out of range.
   *
   * @param int $client_number
   *   The number of the client storage. Default to 0.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function removeStorageClientId(
    int $client_number = 0
  ) :self;

  /**
   * Retrieves the requested storage client.
   *
   * @param int $client_number
   *   The number of the client storage. Default to 0.
   *
   * @return \Drupal\external_entities\StorageClient\StorageClientInterface
   *   The requested storage client plugin.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   Thrown if the storage client plugin could not be retrieved or if the
   *   given client number is out of range (therefore the client can't be
   *   found).
   */
  public function getStorageClient(int $client_number = 0)
  :StorageClientInterface;

  /**
   * Retrieves all the storage clients.
   *
   * @return \Drupal\external_entities\StorageClient\StorageClientInterface[]
   *   The storage client plugins.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   Thrown if a storage client plugin could not be retrieved.
   */
  public function getStorageClients() :array;

  /**
   * Retrieves the configuration of the given storage client plugin.
   *
   * @param int $client_number
   *   The number of the client storage. Default to 0.
   *
   * @return array
   *   An associative array with the storage client configuration. If the
   *   requested client number has no config set, returns an empty array as
   *   default value.
   */
  public function getStorageClientConfig(int $client_number = 0) :array;

  /**
   * Sets the configuration of the given storage client plugin.
   *
   * If the client number is out of range, nothing is set.
   *
   * @param array $storage_client_config
   *   The new configuration for the storage client.
   * @param int $client_number
   *   The number of the client storage. Default to 0.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function setStorageClientConfig(
    array $storage_client_config,
    int $client_number = 0
  ) :self;

  /**
   * Retrieves the administrative notes for the given storage client number.
   *
   * @param int $client_number
   *   The number of the client storage. Default to 0.
   *
   * @return string
   *   The administratives notes of the given storage client. If the requested
   *   client number is not set, returns an empty string as default value.
   */
  public function getStorageClientNotes(int $client_number = 0) :string;

  /**
   * Sets the administrative notes for the given storage client number.
   *
   * If the client number is not 0 and is out of range, nothing is set.
   *
   * @param string $storage_client_notes
   *   The administrative notes for the given storage client.
   * @param int $client_number
   *   The number of the client storage. Default to 0.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function setStorageClientNotes(
    string $storage_client_notes,
    int $client_number = 0
  ) :self;

  /**
   * Retrieves the plugin ID of the data aggregator for this type.
   *
   * @return string
   *   The plugin ID of the data aggregator.
   */
  public function getDataAggregatorId() :string;

  /**
   * Sets the plugin ID of the data aggregator for this type.
   *
   * @param string $data_aggregator_id
   *   The plugin ID of the data aggregator.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function setDataAggregatorId(
    string $data_aggregator_id,
  ) :self;

  /**
   * Returns the data aggregator.
   *
   * If the selected data aggregator plugin id is not available, it falls back
   * to the default data aggregator (group aggregator).
   *
   * @return \Drupal\external_entities\DataAggregator\DataAggregatorInterface
   *   The data aggregator plugin.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   Thrown if nor the selected data aggregator plugin or the default one
   *   could be loaded.
   */
  public function getDataAggregator()
  :DataAggregatorInterface;

  /**
   * Retrieves the configuration of the data aggregator plugin.
   *
   * @return array
   *   An associative array with the data aggregator configuration.
   */
  public function getDataAggregatorConfig() :array;

  /**
   * Sets the configuration of the data aggregator plugin.
   *
   * @param array $data_aggregator_config
   *   The new configuration for the data aggregator.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function setDataAggregatorConfig(
    array $data_aggregator_config
  ) :self;

  /**
   * Retrieves the aggregation settings of the given storage client plugin.
   *
   * @param int $client_number
   *   The number of the client storage.
   *
   * @return array
   *   An associative array with the following structure:
   *   @endcode
   *   [
   *     'groups' => string[] (an array of group prefixes which can be empty),
   *     'group_prefix_strip' => bool (TRUE means the group prefix is stripped),
   *     'id_field' => string (ID field name in storage data),
   *     'merge' => string (one of 'keep', 'over', 'oven', 'ovem' and 'sub'),
   *     'merge_as_member' => string (field name to use to store returned data),
   *     'merge_join' => string (ID field name to use in previous storage data),
   *     'readonly' => bool (TRUE means the storage client is read only),
   *   ],
   *   @code
   */
  public function getStorageClientAggregationConfig(int $client_number) :array;

  /**
   * Sets the aggregation settings for the given storage client plugin.
   *
   * If the client number is out of range, nothing is set.
   *
   * @param array $storage_client_aggr_config
   *   The new aggregation settings for the storage client.
   *   See self::getStorageClientAggregation() return value for details.
   * @param int $client_number
   *   The number of the client storage.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   Returns current instance.
   */
  public function setStorageClientAggregationConfig(
    array $storage_client_aggr_config,
    int $client_number
  ) :self;

  /**
   * Gets the maximum age for these entities persistent cache.
   *
   * @return int
   *   The maximum age in seconds. -1 means the entities are cached permanently,
   *   while 0 means entity caching for this external entity type is disabled.
   */
  public function getPersistentCacheMaxAge() :int;

  /**
   * Gets the ID of the associated content entity type.
   *
   * @return string
   *   The entity type ID.
   */
  public function getDerivedEntityTypeId() :string;

  /**
   * Gets the associated content entity type definition.
   *
   * @return \Drupal\Core\Entity\ContentEntityTypeInterface|null
   *   The entity type definition or NULL if it doesn't exist.
   */
  public function getDerivedEntityType() :ContentEntityTypeInterface|null;

  /**
   * Returns whether the external entity can be annotated.
   *
   * @return bool
   *   TRUE if the entity can be annotated, FALSE otherwise.
   */
  public function isAnnotatable() :bool;

  /**
   * Returns the annotations entity type id.
   *
   * @return string|null
   *   An entity type id or NULL if not annotatable.
   */
  public function getAnnotationEntityTypeId() :string|null;

  /**
   * Returns the annotations bundle.
   *
   * @return string|null
   *   A bundle or NULL if not annotatable.
   */
  public function getAnnotationBundleId() :string|null;

  /**
   * Returns the annotations field name.
   *
   * @return string|null
   *   A field name or NULL if not annotatable.
   */
  public function getAnnotationFieldName() :string|null;

  /**
   * Returns the annotations field.
   *
   * @return \Drupal\Core\Field\FieldDefinitionInterface|null
   *   A field definition or NULL if not annotatable.
   */
  public function getAnnotationField() :FieldDefinitionInterface|null;

  /**
   * Returns if the external entity inherits its annotation entities fields.
   *
   * @return bool
   *   TRUE if fields are inherited, FALSE otherwise.
   */
  public function inheritsAnnotationFields() :bool;

  /**
   * Returns the base path for these external entities.
   *
   * The base path is used to construct the routes these entities live on.
   *
   * @return string
   *   A URL compatible string.
   */
  public function getBasePath() :string;

}
