<?php

namespace Drupal\external_entities\Form;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\ContentEntityType;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\external_entities\DataAggregator\DataAggregatorInterface;
use Drupal\external_entities\Entity\ExternalEntityType;
use Drupal\external_entities\Entity\ExternalEntityTypeInterface;
use Drupal\external_entities\Plugin\ExternalEntities\StorageClient\FileClientInterface;
use Drupal\external_entities\Plugin\ExternalEntities\StorageClient\QueryLanguageClientInterface;
use Drupal\external_entities\Plugin\ExternalEntities\StorageClient\RestClientInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form handler for the external entity type add and edit forms.
 */
class ExternalEntityTypeForm extends EntityForm {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The entity type bundle info service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $entityTypeBundleInfo;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The date formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * The field mapper manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $fieldMapperManager;

  /**
   * The data aggregator manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $dataAggregatorManager;

  /**
   * The external storage client manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $storageClientManager;

  /**
   * Constructs an ExternalEntityTypeForm object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The entity type bundle service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   *   The date formatter service.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $field_mapper_manager
   *   The field mapper manager.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $data_aggregator_manager
   *   The data aggregator manager.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $storage_client_manager
   *   The external storage client manager.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, EntityFieldManagerInterface $entity_field_manager, EntityTypeBundleInfoInterface $entity_type_bundle_info, MessengerInterface $messenger, DateFormatterInterface $date_formatter, PluginManagerInterface $field_mapper_manager, PluginManagerInterface $data_aggregator_manager, PluginManagerInterface $storage_client_manager) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
    $this->messenger = $messenger;
    $this->dateFormatter = $date_formatter;
    $this->fieldMapperManager = $field_mapper_manager;
    $this->dataAggregatorManager = $data_aggregator_manager;
    $this->storageClientManager = $storage_client_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('messenger'),
      $container->get('date.formatter'),
      $container->get('plugin.manager.external_entities.field_mapper'),
      $container->get('plugin.manager.external_entities.data_aggregator'),
      $container->get('plugin.manager.external_entities.storage_client')
    );
  }

  /**
   * Provides an edit external entity type title callback.
   *
   * @param string $entity_type_id
   *   The entity type id.
   *
   * @return string
   *   The title for the entity type edit page.
   */
  public function title($entity_type_id = NULL) {
    $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
    return $entity_type->getLabel();
  }

  /**
   * {@inheritdoc}
   *
   * For external entity types created from config, it is possible to restrict
   * parts of the config form. To do so, you have to set an entry corresponding
   * to the given external entity type into the Drupal state
   * 'external_entities.type.locked' as follow:
   * @code
   *   $locked = \Drupal::state()->get('external_entities.type.locked');
   *   $locked[<your_xntt_type_id>] = [
   *     // Lock external entity editing.
   *     'lock_edit' => TRUE,
   *     // Prevent external entity from being removed.
   *     'lock_delete' => TRUE,
   *     // Disables field management.
   *     'lock_fields' => TRUE,
   *     // Disables field display management (including form display and field
   *     // management).
   *     'lock_display' => TRUE,
   *     // Disables annotation management.
   *     'lock_annotations' => TRUE,
   *     // Limits the number of field mappers to 1.
   *     'lock_field_mapper_count' => 1,
   *     // Sets each field mapper restriction.
   *     'lock_field_mappers' => [
   *       [
   *         // Limits allowed plugins to the given white list (id => TRUE).
   *         // If the list is empty, all plugins are allowed.
   *         'allow_plugins' => [
   *           'simple' => TRUE,
   *         ],
   *         // Config may be visible but not editable.
   *         'lock_config' => TRUE,
   *         // Config is hidden.
   *         'hide_config' => TRUE,
   *       ],
   *     ],
   *     'lock_storage_client_count' => 1,
   *     'lock_storage_clients' => [
   *       [
   *         // Limits allowed plugins to the given white list (id => TRUE).
   *         // If the list is empty, all plugins are allowed.
   *         'allow_plugins' => [
   *           'rest' => TRUE,
   *         ],
   *         // Config may be visible but not editable.
   *         'lock_config' => TRUE,
   *         // Config is hidden.
   *         'hide_config' => TRUE,
   *       ],
   *     ],
   *   ];
   *   Drupal::state()->set('external_entities.type.locked', $locked);
   * @endcode
   */
  public function form(array $form, FormStateInterface $form_state) {
    $external_entity_type = $this->getEntity();

    // Check if external entity type has locks.
    $this->lockStatus = \Drupal::state()->get('external_entities.type.locked');
    if (!empty($this->lockStatus[$external_entity_type->id()]['lock_edit'])) {
      return [
        '#type' => 'markup',
        '#markup' => $this->t('This external entity type is not editable.'),
      ];
    }

    $form = parent::form($form, $form_state);

    if ($this->operation == 'add') {
      $form['#title'] = $this->t('Add external entity type');
    }
    else {
      $form['#title'] = $this->t('Edit %label external entity type', ['%label' => $external_entity_type->label()]);
    }

    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Name'),
      '#maxlength' => 255,
      '#default_value' => $external_entity_type->label(),
      '#description' => $this->t('The human-readable name of this external entity type. This name must be unique.'),
      '#required' => TRUE,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $external_entity_type->id(),
      '#machine_name' => [
        'exists' => [$this, 'exists'],
      ],
      '#maxlength' => EntityTypeInterface::ID_MAX_LENGTH,
      '#disabled' => !$external_entity_type->isNew(),
    ];

    $form['label_plural'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Name (plural)'),
      '#maxlength' => 255,
      '#default_value' => $external_entity_type->getPluralLabel(),
      '#description' => $this->t('The plural human-readable name of this external entity type.'),
      '#required' => TRUE,
    ];

    $form['description'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Description'),
      '#description' => $this->t("Description of the external entity type."),
      '#default_value' => $external_entity_type->getDescription(),
    ];

    $form['read_only'] = [
      '#title' => $this->t('Read only'),
      '#type' => 'checkbox',
      '#default_value' => $external_entity_type->isReadOnly(),
      '#description' => $this->t('Whether or not this external entity type is read only.'),
    ];

    $form['debug_level'] = [
      '#title' => $this->t('Debug mode'),
      '#type' => 'select',
      '#options' => [
        0 => $this->t('Disabled'),
        1 => $this->t('Log events'),
        2 => $this->t('Log events with details'),
        8 => $this->t('Log events and do not perform altering operations'),
      ],
      '#default_value' => $external_entity_type->getDebugLevel() ?? 0,
      '#description' => $this->t('Enables debugging: it also sets data aggregator, storage client and field mapper to debug mode.'),
    ];

    $form['generate_aliases'] = [
      '#title' => $this->t('Automatically generate aliases'),
      '#type' => 'checkbox',
      '#default_value' => $external_entity_type->automaticallyGenerateAliases(),
      '#description' => $this->t('Whether or not to automatically generate aliases for this external entity type.'),
    ];

    $form['additional_settings'] = [
      '#type' => 'vertical_tabs',
    ];

    $form['field_mapper'] = [
      '#type' => 'details',
      '#title' => $this->t('Field mapping'),
      '#group' => 'additional_settings',
      '#open' => TRUE,
    ];

    try {
      $this->buildFieldMapperSelectForm($form, $form_state);
      $this->buildFieldMapperConfigForm($form, $form_state);
    }
    catch (PluginException $e) {
      $form['field_mapper']['field_mapper_config'] = [
        '#type' => 'item',
        '#markup' => $this->t(
          'WARNING: Failed to load a field mapper plugin!'
        ),
      ];
      $this->logger('external_entities')->error(
        'Failed to load a field mapper plugin: '
        . $e
      );
    }

    // Storage clients.
    $storage_count =
      $form_state->get('storage_count')
      ?? $external_entity_type->getStorageClientCount();

    // We must have at least one storage.
    if (empty($storage_count)) {
      $storage_count = 1;
    }
    $form_state->set('storage_count', $storage_count);
    // Get storage client configs: try from form state first and then from
    // xntt config.
    $storage_settings = $form_state->get('storage_settings');
    if ((!isset($storage_settings) || empty($storage_settings))) {
      $storage_settings = [];
      if (!$external_entity_type->isNew()) {
        for ($i = 0; $i < $storage_count; ++$i) {
          $storage_settings[$i] = [
            'id' => $external_entity_type->getStorageClientId($i),
            'config' => $external_entity_type->getStorageClientConfig($i),
          ];
        }
      }
      else {
        for ($i = 0; $i < $storage_count; ++$i) {
          $storage_settings[$i] = [
            'id' => '',
            'config' => [],
          ];
        }
      }
      $form_state->set('storage_settings', $storage_settings);
    }
    // Get current aggregations settings.
    $aggr_settings = $form_state->get('aggr_settings');
    if ((!isset($aggr_settings) || empty($aggr_settings))) {
      $aggr_settings = [];
      if (!$external_entity_type->isNew()) {
        $aggr_settings = [
          'id' => $external_entity_type->getDataAggregatorId(),
          'config' => $external_entity_type->getDataAggregatorConfig(),
        ];
      }
      else {
        $aggr_settings = [
          'id' => ExternalEntityType::DEFAULT_DATA_AGGREGATOR,
          'config' => [
            'global_config' => [],
            'storage_client_config' => [
              [],
              [],
            ],
          ],
        ];
      }
      $form_state->set('aggr_settings', $aggr_settings);
    }

    // Check if we use multiple storage clients.
    $is_multi =
      $form_state->getValue(['storages_tab', 'storages', 'multi'], FALSE)
      || (1 < $storage_count);

    $form['storages_tab'] = [
      '#type' => 'details',
      '#title' => $this->t('Storage'),
      '#group' => 'additional_settings',
      '#open' => FALSE,
    ];
    $form['storages_tab']['storages'] = [
      '#type' => 'container',
      '#attributes' => [
        'id' => 'external-entities-storage-clients-form',
      ],
    ];
    $form['storages_tab']['storages']['aggr'] = [];

    $form['storages_tab']['storages']['multi'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Combine multiple storage clients'),
      '#default_value' => $is_multi,
      '#disabled' => (1 < $storage_count),
      '#ajax' => [
        'callback' => [get_class($this), 'buildAjaxStorageClientsForm'],
        'wrapper' => 'external-entities-storage-clients-form',
        'method' => 'replace',
        'effect' => 'fade',
      ],
    ];
    // Check for plugin restrictions.
    if (!empty($this->lockStatus[$external_entity_type->id()]['lock_data_aggregator']['lock_config'])) {
      $form['storages_tab']['storages']['multi']['#disabled'] = TRUE;
    }
    if (!empty($this->lockStatus[$external_entity_type->id()]['lock_data_aggregator']['hide_config'])) {
      $form['storages_tab']['storages']['multi']['#type'] = 'hidden';
    }

    if ($is_multi) {
      $form['storages_tab']['storages']['multi_info'] = [
        '#type' => 'item',
        '#markup' => $this->t(
          'You can specify one or more storage client. Their order is important: the first one (of a group) is loaded first and will be used as the reference (for its group). Each group reference is considered holding all the records of the group. Other records provided by the following storage clients will be discarded.
        '),
      ];
      $this->buildDataAggregatorSelectForm($form, $form_state);
      $this->buildGlobalAggregationForm($form, $form_state);
    }

    for ($client_number = 0; $client_number < $storage_count; ++$client_number) {
      if ($is_multi) {
        $form['storages_tab']['storages'][$client_number] = [
          '#type' => 'details',
          '#title' => $this->t(
            'Storage client #@index',
            ['@index' => $client_number + 1]
          ),
          '#open' => TRUE,
          '#attributes' => [
            'id' => 'external-entities-storage-client-' . $client_number . '-config-form',
          ],
        ];

        // Data aggregation settings.
        $form['storages_tab']['storages'][$client_number]['aggr'] = [
          '#type' => 'fieldset',
          '#title' => $this->t('Data aggregation settings'),
          '#open' => TRUE,
        ];
        $this->buildStorageClientAggregationForm($form, $form_state, $client_number);
      }
      else {
        $form['storages_tab']['storages'][$client_number] = [
          '#type' => 'container',
          '#attributes' => [
            'id' => 'external-entities-storage-client-' . $client_number . '-config-form',
          ],
        ];
        $form['storages_tab']['storages'][$client_number]['aggr'] = [];
      }
      try {
        $this->buildStorageClientSelectForm($form, $form_state, $client_number, $is_multi);
        $this->buildStorageClientConfigForm($form, $form_state, $client_number, $is_multi);
      }
      catch (PluginException $e) {
        $form['storages_tab']['storages'][$client_number] = [
          '#type' => 'item',
          '#markup' => $this->t(
            'WARNING: Failed to load a client storage plugin!'
          ),
        ];
        $this->logger('external_entities')->error(
          'Failed to load a client storage plugin: '
          . $e
        );
      }
      if (1 < $storage_count) {
        $form['storages_tab']['storages'][$client_number]['remove_storage'] = [
          '#type' => 'submit',
          '#value' => $this->t(
            'Remove storage client @number',
            [
              '@number' => '#' . ($client_number + 1),
            ]
          ),
          '#name' => 'remove_storage',
          '#attributes' => ['data-remove' => $client_number],
          '#ajax' => [
            'callback' => [get_class($this), 'buildAjaxStorageClientsForm'],
            'wrapper' => 'external-entities-storage-clients-form',
            'method' => 'replace',
            'effect' => 'fade',
          ],
        ];
      }
    }

    // Append button to add storage clients.
    if ($is_multi) {
      $form['storages_tab']['storages']['add_storage'] = [
        '#type' => 'submit',
        '#value' => $this->t('Add a storage'),
        '#name' => 'add_storage',
        '#ajax' => [
          'callback' => [get_class($this), 'buildAjaxStorageClientsForm'],
          'wrapper' => 'external-entities-storage-clients-form',
          'method' => 'replace',
          'effect' => 'fade',
        ],
      ];
    }

    $form['caching'] = [
      '#type' => 'details',
      '#title' => $this->t('Caching'),
      '#group' => 'additional_settings',
      '#open' => FALSE,
    ];

    $period = [
      0,
      60,
      180,
      300,
      600,
      900,
      1800,
      2700,
      3600,
      10800,
      21600,
      32400,
      43200,
      86400,
    ];
    $period = array_map([$this->dateFormatter, 'formatInterval'], array_combine($period, $period));
    $period[ExternalEntityType::CACHE_DISABLED] = '<' . $this->t('no caching') . '>';
    $period[Cache::PERMANENT] = $this->t('Permanent');
    $persistent_cache_max_age = $form_state->getValue(['caching', 'entity_cache_max_age'], ExternalEntityType::CACHE_DISABLED)
        ?: $external_entity_type->getPersistentCacheMaxAge();
    $form['caching']['persistent_cache_max_age'] = [
      '#type' => 'select',
      '#title' => $this->t('Persistent cache maximum age'),
      '#description' => $this->t('The maximum time the external entity can be persistently cached.'),
      '#options' => $period,
      '#default_value' => $persistent_cache_max_age,
    ];

    $form['annotations'] = [
      '#type' => 'details',
      '#title' => $this->t('Annotations'),
      '#group' => 'additional_settings',
      '#open' => FALSE,
    ];

    if (!empty($this->lockStatus[$this->getEntity()->id()]['lock_annotations'])) {
      $form['annotations']['#disabled'] = TRUE;
    }

    $annotations_enable = $form_state->getValue(['annotations', 'annotations_enable']);
    if ($annotations_enable === NULL) {
      $annotations_enable = $external_entity_type->isAnnotatable();
    }
    $form['annotations']['annotations_enable'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enable annotations on this external entity type'),
      '#description' => $this->t('Annotations allow to enrich external entities with locally stored data (such as Drupal fields). This is achieved by linking local entities with external entities using an entity reference field (on the local entity). Before enabling this option, you have to make sure an entity reference field (to the external entity type) is available on the local entity type and/or bundle.'),
      '#default_value' => $annotations_enable,
      '#ajax' => [
        'callback' => [$this, 'refreshAnnotationSettings'],
        'wrapper' => 'annotation-settings-wrapper',
      ],
    ];

    $form['annotations']['annotation_settings'] = [
      '#type' => 'container',
      '#prefix' => '<div id="annotation-settings-wrapper">',
      '#suffix' => '</div>',
    ];

    if ($annotations_enable) {
      $annotation_entity_type_id_options = $this->getAnnotationEntityTypeIdOptions();
      $annotation_entity_type_id = $form_state->getValue([
        'annotations',
        'annotation_settings',
        'annotation_entity_type_id',
      ]) ?: $external_entity_type->getAnnotationEntityTypeId();
      $form['annotations']['annotation_settings']['annotation_entity_type_id'] = [
        '#type' => 'select',
        '#title' => $this->t('Entity type'),
        '#description' => $this->t('Entity type used for annotations.'),
        '#options' => $annotation_entity_type_id_options,
        '#ajax' => [
          'callback' => [$this, 'refreshAnnotationBundleOptions'],
          'wrapper' => 'annotation-bundle-id-wrapper',
        ],
        '#default_value' => $annotation_entity_type_id,
        '#required' => TRUE,
      ];

      $annotation_bundle_id = $form_state->getValue([
        'annotations',
        'annotation_settings',
        'annotation_bundle_id',
      ]) ?: $external_entity_type->getAnnotationBundleId();
      $form['annotations']['annotation_settings']['annotation_bundle_id'] = [
        '#prefix' => '<div id="annotation-bundle-id-wrapper">',
        '#type' => 'select',
        '#title' => $this->t('Bundle'),
        '#description' => $this->t('Bundle used for annotations.'),
        '#options' => $this->getAnnotationBundleIdOptions($annotation_entity_type_id),
        '#ajax' => [
          'callback' => [$this, 'refreshAnnotationFieldOptions'],
          'wrapper' => 'annotation-field-config-id-wrapper',
        ],
        '#default_value' => $annotation_bundle_id,
        '#disabled' => !$annotation_entity_type_id,
        '#required' => TRUE,
      ];

      $annotation_field_name = $form_state->getValue([
        'annotations',
        'annotation_settings',
        'annotation_field_name',
      ]) ?: $external_entity_type->getAnnotationFieldName();
      $annotation_field_options = $this->getAnnotationFieldOptions($annotation_entity_type_id, $annotation_bundle_id);
      $annotation_field_default_value = !empty($annotation_field_options[$annotation_field_name])
        ? $annotation_field_options[$annotation_field_name]
        : NULL;
      $form['annotations']['annotation_settings']['annotation_field_name'] = [
        '#prefix' => '<div id="annotation-field-config-id-wrapper">',
        '#suffix' => '</div></div>',
        '#type' => 'select',
        '#title' => $this->t('Entity reference field'),
        '#description' => $this->t('The entity reference field on the annotation entity which references the external entity.'),
        '#options' => $annotation_field_options,
        '#default_value' => $annotation_field_default_value,
        '#disabled' => !$annotation_entity_type_id || !$annotation_bundle_id,
        '#required' => TRUE,
      ];

      $inherits_annotation_fields = $form_state->getValue([
        'annotations',
        'annotation_settings',
        'inherits_annotation_fields',
      ], FALSE) ?: $external_entity_type->inheritsAnnotationFields();
      $form['annotations']['annotation_settings']['inherits_annotation_fields'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Inherit fields'),
        '#description' => $this->t('When enabled, fields and their values are automatically transferred from the annotation to the external entity. Inherited fields are regular entity fields and can be used as such.'),
        '#default_value' => $inherits_annotation_fields,
      ];
    }

    $form['#tree'] = TRUE;

    return $form;
  }

  /**
   * AJAX callback which refreshes the annotation settings.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   *
   * @return array
   *   The form structure.
   */
  public function refreshAnnotationSettings(array $form) {
    return $form['annotations']['annotation_settings'] ?: [];
  }

  /**
   * AJAX callback which refreshes the annotation bundle options.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   *
   * @return array
   *   The form structure.
   */
  public function refreshAnnotationBundleOptions(array $form) {
    return [
      'annotations' => [
        'annotation_settings' => [
          'annotation_bundle_id' => $form['annotations']['annotation_settings']['annotation_bundle_id'] ?: [],
          'annotation_field_name' => $form['annotations']['annotation_settings']['annotation_field_name'] ?: [],
        ],
      ],
    ];
  }

  /**
   * AJAX callback which refreshes the annotation field options.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   *
   * @return array
   *   The form structure.
   */
  public function refreshAnnotationFieldOptions(array $form) {
    return $form['annotations']['annotation_settings']['annotation_field_name'] ?: [];
  }

  /**
   * Builds the field mapper selection configuration.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   If a field mapper plug-in cannot be loaded.
   */
  public function buildFieldMapperSelectForm(array &$form, FormStateInterface $form_state) {
    $external_entity_type = $this->getEntity();
    $field_mapper_options = [];
    $field_mapper_descriptions = [];

    $field_mappers = $this->fieldMapperManager->getDefinitions();
    foreach ($field_mappers as $field_mapper_id => $definition) {
      // Check for plugin restrictions.
      if (!empty($this->lockStatus[$external_entity_type->id()]['lock_field_mappers'][0]['allow_plugins'])
        && empty($this->lockStatus[$external_entity_type->id()]['lock_field_mappers'][0]['allow_plugins'][$field_mapper_id])
      ) {
        continue;
      }

      if ($field_mapper_id === $external_entity_type->getFieldMapperId()) {
        // This mapper was previously used; init it with saved config.
        $config = $external_entity_type->getFieldMapperConfig();
      }
      else {
        // This mapper is not yet used; init it with a blank config.
        $config = [];
      }

      // Allow the mapper to call back into the entity type (e.g., to fetch
      // additional data like field lists from the remote service).
      $config[ExternalEntityTypeInterface::XNTT_TYPE_PROP] = $external_entity_type;

      $field_mapper = $this->fieldMapperManager->createInstance($field_mapper_id, $config);
      $field_mapper_options[$field_mapper_id] = $field_mapper->getLabel();
      $field_mapper_descriptions[$field_mapper_id]['#description'] = $field_mapper->getDescription();
    }
    asort($field_mapper_options, SORT_NATURAL | SORT_FLAG_CASE);

    if ($field_mapper_options) {
      $form['field_mapper']['field_mapper_id'] = [
        '#type' => 'radios',
        '#title' => $this->t('Field mapper'),
        '#description' => (1 < count($field_mapper_options))
          ? $this->t('Choose a field mapper to use for this type.')
          : '',
        '#options' => $field_mapper_options,
        '#default_value' => $external_entity_type->getFieldMapperId()
          ?: ExternalEntityType::DEFAULT_FIELD_MAPPER,
        '#required' => TRUE,
        '#ajax' => [
          'callback' => [get_class($this), 'buildAjaxFieldMapperConfigForm'],
          'wrapper' => 'external-entities-field-mapper-config-form',
          'method' => 'replace',
          'effect' => 'fade',
        ],
      ];
      $form['field_mapper']['field_mapper_id'] += $field_mapper_descriptions;
    }
  }

  /**
   * Builds the field mapper-specific configuration form.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   If a field mapper plugin fails to load.
   */
  public function buildFieldMapperConfigForm(array &$form, FormStateInterface $form_state) {
    $external_entity_type = $this->getEntity();
    $form['field_mapper']['field_mapper_config'] = [];

    $field_mapper_id =
      $form_state->getValue(['field_mapper', 'field_mapper_id']) ?:
        $external_entity_type->getFieldMapperId();

    if ($field_mapper_id) {
      // Check for plugin restrictions.
      if (!empty($this->lockStatus[$external_entity_type->id()]['lock_field_mappers'][0]['lock_config'])) {
        $form['field_mapper']['#disabled'] = TRUE;
      }

      if (!$external_entity_type->isNew() &&
          $field_mapper_id === $external_entity_type->getFieldMapperId()) {
        // This is the currently selected field mapper; use the loaded instance.
        $field_mapper = $external_entity_type->getFieldMapper();
      }
      else {
        $mapper_config = [
          // Allow the mapper to call back into the entity type (e.g., to fetch
          // additional data like field lists from the remote service).
          ExternalEntityTypeInterface::XNTT_TYPE_PROP => $external_entity_type,
        ];

        $field_mapper = $this->fieldMapperManager->createInstance(
          $field_mapper_id,
          $mapper_config
        );
      }

      if ($field_mapper instanceof PluginFormInterface) {
        if ($external_entity_type->isNew()) {
          $form['field_mapper']['field_mapper_config']['message_new'] = [
            '#type' => 'item',
            '#markup' => $this->t('You must save this external entity type before being able to map fields.'),
          ];
        }
        else {
          // Attach the field mapper plugin configuration form.
          $field_mapper_form_state = SubformState::createForSubform($form['field_mapper']['field_mapper_config'], $form, $form_state);
          $form['field_mapper']['field_mapper_config'] = $field_mapper->buildConfigurationForm($form['field_mapper']['field_mapper_config'], $field_mapper_form_state);

          // Modify the field mapper plugin configuration container element.
          $form['field_mapper']['field_mapper_config']['#type'] = 'fieldset';
          $form['field_mapper']['field_mapper_config']['#title'] = $this->t('Configure %plugin field mapper', ['%plugin' => $field_mapper->getLabel()]);
          $form['field_mapper']['field_mapper_config']['#open'] = TRUE;
        }
      }
    }

    $form['field_mapper']['field_mapper_config'] += [
      '#type' => 'container',
    ];
    $form['field_mapper']['field_mapper_config']['#attributes']['id'] = 'external-entities-field-mapper-config-form';

    // Check for plugin restrictions.
    if (!empty($this->lockStatus[$external_entity_type->id()]['lock_field_mappers'][0]['hide_config'])) {
      $form['field_mapper']['#type'] = 'hidden';
    }
  }

  /**
   * Handles switching the selected field mapper plugin.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The part of the form to return as AJAX.
   */
  public static function buildAjaxFieldMapperConfigForm(array $form, FormStateInterface $form_state) {
    // The work is already done in form(), where we rebuild the entity according
    // to the current form values and then create the storage client
    // configuration form based on that. So we just need to return the relevant
    // part of the form
    // here.
    return $form['field_mapper']['field_mapper_config'];
  }

  /**
   * Builds the data aggregator selection configuration.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   If a data aggregator plug-in cannot be loaded.
   */
  public function buildDataAggregatorSelectForm(array &$form, FormStateInterface $form_state) {
    $external_entity_type = $this->getEntity();
    $data_aggregator_options = [];
    $data_aggregator_descriptions = [];

    $data_aggregators = $this->dataAggregatorManager->getDefinitions();
    if (1 >= count($data_aggregators)) {
      return;
    }
    foreach ($data_aggregators as $data_aggregator_id => $definition) {
      // Check for plugin restrictions.
      if (!empty($this->lockStatus[$external_entity_type->id()]['lock_data_aggregators']['allow_plugins'])
        && empty($this->lockStatus[$external_entity_type->id()]['lock_data_aggregators']['allow_plugins'][$data_aggregator_id])
      ) {
        continue;
      }

      if ($data_aggregator_id === $external_entity_type->getDataAggregatorId()) {
        // This aggregator was previously used; init it with saved config.
        $config = $external_entity_type->getDataAggregatorConfig();
      }
      else {
        // This aggregator is not yet used; init it with a blank config.
        $config = [];
      }

      // Allow the aggregator to call back into the entity type to get storage
      // client settings.
      $config[ExternalEntityTypeInterface::XNTT_TYPE_PROP] = $external_entity_type;

      $data_aggregator = $this->dataAggregatorManager->createInstance($data_aggregator_id, $config);
      $data_aggregator_options[$data_aggregator_id] = $data_aggregator->getLabel();
      $data_aggregator_descriptions[$data_aggregator_id]['#description'] = $data_aggregator->getDescription();
    }
    asort($data_aggregator_options, SORT_NATURAL | SORT_FLAG_CASE);

    if ($data_aggregator_options) {
      $form['storages_tab']['storages']['data_aggregator_id'] = [
        '#type' => 'radios',
        '#title' => $this->t('Data aggregator'),
        '#description' => $this->t('Choose a data aggregator to use for this type.'),
        '#options' => $data_aggregator_options,
        '#default_value' => $external_entity_type->getDataAggregatorId() ?: 'group',
        '#required' => TRUE,
        '#ajax' => [
          'callback' => [get_class($this), 'buildAjaxDataAggregatorConfigForm'],
          'wrapper' => 'external-entities-storage-clients-form',
          'method' => 'replace',
          'effect' => 'fade',
        ],
      ];
      $form['storages_tab']['storages']['data_aggregator_id'] += $data_aggregator_descriptions;
    }
  }

  /**
   * Builds the global aggregation configuration form.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public function buildGlobalAggregationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    $external_entity_type = $this->getEntity();
    $aggr_settings =
      ($form_state->get('aggr_settings') ?? [])
      + ['config' => []];
    $form['storages_tab']['storages']['aggr'] = [];

    $xntt_type = $this->getEntity();
    $data_aggregator = $xntt_type->getDataAggregator();
    if ($data_aggregator && $data_aggregator instanceof DataAggregatorInterface) {
      $data_aggregator->setConfiguration(
        // Allow the data aggregator to call back into the entity type to fetch
        // storage client settings.
        $aggr_settings['config']
        + [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $xntt_type],
      );
      // Attach the data aggregator plugin configuration form.
      $data_aggregator_form_state = SubformState::createForSubform(
        $form['storages_tab']['storages']['aggr'],
        $form,
        $form_state
      );
      $form['storages_tab']['storages']['aggr'] =
        $data_aggregator->buildGlobalAggregationForm(
          $form['storages_tab']['storages']['aggr'],
          $data_aggregator_form_state
        );

      // Check for plugin restrictions.
      if (!empty($this->lockStatus[$xntt_type->id()]['lock_data_aggregator']['lock_config'])) {
        $form['storages_tab']['storages']['aggr']['#disabled'] = TRUE;
      }
      if (!empty($this->lockStatus[$xntt_type->id()]['lock_data_aggregator']['hide_config'])) {
        $form['storages_tab']['storages']['aggr']['#type'] = 'hidden';
      }
    }
  }

  /**
   * Builds the storage client-specific aggregation configuration form.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param int $client_number
   *   Storage client number for this external entity.
   */
  public function buildStorageClientAggregationForm(
    array &$form,
    FormStateInterface $form_state,
    int $client_number
  ) {
    $aggr_settings =
      ($form_state->get('aggr_settings') ?? [])
      + ['config' => []];
    $form['storages_tab']['storages'][$client_number]['aggr'] ??= [];

    $xntt_type = $this->getEntity();
    $data_aggregator = $xntt_type->getDataAggregator();
    if ($data_aggregator && $data_aggregator instanceof DataAggregatorInterface) {
      $data_aggregator->setConfiguration(
        // Allow the data aggregator to call back into the entity type to fetch
        // storage client settings.
        $aggr_settings['config']
        + [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $xntt_type],
      );
      // Attach the data aggregator plugin configuration form for the given
      // storage client.
      $data_aggregator_form_state = SubformState::createForSubform(
        $form['storages_tab']['storages'][$client_number]['aggr'],
        $form,
        $form_state
      );
      $form['storages_tab']['storages'][$client_number]['aggr'] =
        $data_aggregator->buildStorageClientAggregationForm(
          $form['storages_tab']['storages'][$client_number]['aggr'],
          $data_aggregator_form_state,
          $client_number
        );

      // Modify the storage client plugin configuration container element.
      $form['storages_tab']['storages'][$client_number]['aggr']['#type'] =
        'fieldset';
      $form['storages_tab']['storages'][$client_number]['aggr']['#title'] =
        $this->t('Data aggregation settings');
      $form['storages_tab']['storages'][$client_number]['aggr']['#open'] =
        TRUE;
      // Check for plugin restrictions.
      if (!empty($this->lockStatus[$xntt_type->id()]['lock_data_aggregator']['lock_config'])) {
        $form['storages_tab']['storages'][$client_number]['aggr']['#disabled'] = TRUE;
      }
      if (!empty($this->lockStatus[$xntt_type->id()]['lock_data_aggregator']['hide_config'])) {
        $form['storages_tab']['storages'][$client_number]['aggr']['#type'] = 'hidden';
      }
    }
    else {
      // If no plugin form was provided, use a default empty container.
      $form['storages_tab']['storages'][$client_number]['aggr'] += [
        '#type' => 'container',
      ];
    }
  }

  /**
   * Handles switching the selected data aggregator plugin.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The part of the form to return as AJAX.
   */
  public static function buildAjaxDataAggregatorConfigForm(array $form, FormStateInterface $form_state) {
    // The work is already done in form(), where we rebuild the entity according
    // to the current form values and then create the storage client
    // configuration form based on that. So we just need to return the relevant
    // part of the form
    // here.
    return $form['storages_tab']['storages'];
  }

  /**
   * Builds the storage client selection configuration.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param int $client_number
   *   Storage client number for this external entity.
   * @param bool $is_compact
   *   Display form in a compact mode..
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   If a storage client plug-in cannot be loaded.
   */
  public function buildStorageClientSelectForm(
    array &$form,
    FormStateInterface $form_state,
    int $client_number = 0,
    bool $is_compact = FALSE
  ) {
    $storage_settings = $form_state->get('storage_settings');
    $external_entity_type = $this->getEntity();
    $storage_clients = $this->storageClientManager->getDefinitions();
    $storage_client_options = [];
    $storage_client_descriptions = [];

    $rest_group = '' . $this->t('Rest clients');
    $files_group = '' . $this->t('File clients');
    $ql_group = '' . $this->t('Query Language clients');
    $others_group = '' . $this->t('Other clients');
    foreach ($storage_clients as $storage_client_id => $definition) {
      // Check for plugin restrictions.
      if (!empty($this->lockStatus[$external_entity_type->id()]['lock_storage_clients'][$client_number]['allow_plugins'])
        && empty($this->lockStatus[$external_entity_type->id()]['lock_storage_clients'][$client_number]['allow_plugins'][$storage_client_id])
      ) {
        continue;
      }

      if ($storage_client_id === $storage_settings[$client_number]['id']) {
        // This storage client was previously used; init it with saved config
        // and allow the storage client to call back into the entity type (e.g.,
        // to fetch additional data or settings).
        $config =
          $storage_settings[$client_number]['config']
          + [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $external_entity_type];
      }
      else {
        // This storage client is not yet used; init it with a blank config but
        // allow the storage client to call back into the entity type (e.g., to
        // fetch additional data or settings).
        $config = [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $external_entity_type];
      }

      $storage_client = $this
        ->storageClientManager
        ->createInstance($storage_client_id, $config);
      if ($is_compact) {
        // @todo Use client interfaces.
        if ($storage_client instanceof RestClientInterface) {
          $storage_client_options[$rest_group][$storage_client_id] = $storage_client->getLabel();
        }
        elseif ($storage_client instanceof FileClientInterface) {
          $storage_client_options[$files_group][$storage_client_id] = $storage_client->getLabel();
        }
        elseif ($storage_client instanceof QueryLanguageClientInterface) {
          $storage_client_options[$ql_group][$storage_client_id] = $storage_client->getLabel();
        }
        else {
          $storage_client_options[$others_group][$storage_client_id] = $storage_client->getLabel();
        }
      }
      else {
        $storage_client_options[$storage_client_id] = $storage_client->getLabel();
        $storage_client_descriptions[$storage_client_id]['#description'] =
          $storage_client->getDescription();
      }
    }
    if ($is_compact) {
      // Sort categories first.
      uksort(
        $storage_client_options,
        function ($a, $b) use ($storage_client_options, $rest_group, $files_group, $ql_group) {
          // Rest first.
          if ($a == $rest_group) {
            return -1;
          }
          elseif ($b == $rest_group) {
            return 1;
          }
          // Then Files.
          if ($a == $files_group) {
            return -1;
          }
          elseif ($b == $files_group) {
            return 1;
          }
          // Then QL.
          if ($a == $ql_group) {
            return -1;
          }
          elseif ($b == $ql_group) {
            return 1;
          }
          // Then others.
          if ($a == $others_group) {
            return -1;
          }
          elseif ($b == $others_group) {
            return 1;
          }
          // Then the uncategorized items.
          return strcasecmp(
            $storage_client_options[$a],
            $storage_client_options[$b]
          );
        }
      );
      // Sort items in each category.
      if (!empty($storage_client_options[$rest_group])) {
        asort($storage_client_options[$rest_group], SORT_NATURAL | SORT_FLAG_CASE);
      }
      if (!empty($storage_client_options[$files_group])) {
        asort($storage_client_options[$files_group], SORT_NATURAL | SORT_FLAG_CASE);
      }
      if (!empty($storage_client_options[$ql_group])) {
        asort($storage_client_options[$ql_group], SORT_NATURAL | SORT_FLAG_CASE);
      }
      if (!empty($storage_client_options[$others_group])) {
        asort($storage_client_options[$others_group], SORT_NATURAL | SORT_FLAG_CASE);
      }
    }
    else {
      asort($storage_client_options, SORT_NATURAL | SORT_FLAG_CASE);
    }

    if ($storage_client_options) {
      $form['storages_tab']['storages'][$client_number]['id'] = [
        '#type' => $is_compact ? 'select' : 'radios',
        '#title' => $this->t('Storage client'),
        '#description' => (1 < count($storage_client_options))
          ? $this->t('Choose a storage client to use, then configure it below.')
          : '',
        '#options' => $storage_client_options,
        // '#empty_option' only used when #type is 'select'.
        // We enforce the empty options because of issue #3180011.
        '#empty_option' => $this->t('- Select -'),
        '#default_value' => ($storage_settings[$client_number]['id'] ?? '')
          ?: ExternalEntityType::DEFAULT_STORAGE_CLIENT,
        '#required' => TRUE,
        '#attributes' => [
          'data-client' => $client_number,
          'autocomplete' => 'off',
        ],
        '#ajax' => [
          'callback' => [get_class($this), 'buildAjaxStorageClientsForm'],
          'wrapper' => 'external-entities-storage-client-' . $client_number . '-config-form',
          'method' => 'replace',
          'effect' => 'fade',
        ],
      ];
      if (!$is_compact) {
        $form['storages_tab']['storages'][$client_number]['id'] +=
          $storage_client_descriptions;
      }
    }
  }

  /**
   * Builds the storage client-specific configuration form.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param int $client_number
   *   Storage client number for this external entity.
   * @param bool $is_compact
   *   Display form in a compact mode..
   */
  public function buildStorageClientConfigForm(
    array &$form,
    FormStateInterface $form_state,
    int $client_number = 0,
    bool $is_compact = FALSE
  ) {
    $external_entity_type = $this->getEntity();
    $storage_settings = $form_state->get('storage_settings');
    $form['storages_tab']['storages'][$client_number]['config'] = [];
    $storage_client_id =
      ($storage_settings[$client_number]['id'] ?? '')
      ?: ExternalEntityType::DEFAULT_STORAGE_CLIENT;
    if (!empty($storage_client_id)) {
      // Check for plugin restrictions.
      if (!empty($this->lockStatus[$external_entity_type->id()]['lock_storage_clients'][$client_number]['lock_config'])) {
        $form['storages_tab']['#disabled'] = TRUE;
      }

      $storage_client = $this
        ->storageClientManager
        ->createInstance($storage_client_id);
      if ($storage_client && $storage_client instanceof PluginFormInterface) {
        $storage_client->setConfiguration(
          // Allow the storage client to call back into the entity type (e.g.,
          // to fetch additional data or settings).
          $storage_settings[$client_number]['config']
          + [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $this->getEntity()],
        );
        // Attach the storage client plugin configuration form.
        $storage_client_form_state = SubformState::createForSubform(
          $form['storages_tab']['storages'][$client_number]['config'],
          $form,
          $form_state
        );
        $form['storages_tab']['storages'][$client_number]['config'] =
          $storage_client->buildConfigurationForm(
            $form['storages_tab']['storages'][$client_number]['config'],
            $storage_client_form_state
          );

        // Modify the storage client plugin configuration container element.
        $form['storages_tab']['storages'][$client_number]['config']['#type'] =
          'fieldset';
        $form['storages_tab']['storages'][$client_number]['config']['#title'] =
          $this->t(
            'Configure %plugin storage client',
            ['%plugin' => $storage_client->getLabel()]
          );
        $form['storages_tab']['storages'][$client_number]['config']['#open'] =
          TRUE;
      }
    }
    // If no plugin form was provided, use a default empty container.
    $form['storages_tab']['storages'][$client_number]['config'] += [
      '#type' => 'container',
    ];
    // Check for plugin restrictions.
    if (!empty($this->lockStatus[$external_entity_type->id()]['lock_storage_clients'][$client_number]['hide_config'])) {
      $form['storages']['#type'] = 'hidden';
    }
  }

  /**
   * Handles storage client form part update.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The part of the form to return as AJAX.
   */
  public static function buildAjaxStorageClientsForm(array $form, FormStateInterface $form_state) {
    // The work is already done in form(), where we rebuild the form according
    // to selection. So we just need to return the relevant part of the
    // form here.
    // However, we can narrow the parts to send if is is related to a given
    // client.
    $trigger = $form_state->getTriggeringElement();
    $client_number = $trigger['#attributes']['data-client'] ?? NULL;
    if (isset($client_number)) {
      $subform = $form['storages_tab']['storages'][$client_number];
    }
    else {
      $subform = $form['storages_tab']['storages'];
    }
    // @todo Clear site messages.
    return $subform;
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    if (!empty($this->lockStatus[$this->getEntity()->id()]['lock_edit'])) {
      // Not editable.
      return [];
    }

    $actions = parent::actions($form, $form_state);
    $actions['submit']['#value'] = $this->t('Save external entity type');

    if (!$this->getEntity()->isNew()) {
      $actions['submit2'] = $actions['submit'];
      $actions['submit2']['#value'] = $this->t('Save & edit');
      $actions['submit2']['#name'] = 'save_and_edit';
    }

    if (!empty($this->lockStatus[$this->getEntity()->id()]['lock_delete'])) {
      // Not deletable.
      unset($actions['delete']);
    }

    return $actions;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $external_entity_type = $this->getEntity();
    // Check for Ajax events.
    $rebuild = FALSE;
    if ($trigger = $form_state->getTriggeringElement()) {
      switch ($trigger['#name']) {
        case 'add_storage':
          $storage_count = $form_state->get('storage_count') + 1;
          $form_state->set('storage_count', $storage_count);
          $storage_settings = $form_state->get('storage_settings');
          $storage_settings[] = [
            'id' => '',
            'config' => [],
          ];
          $form_state->set('storage_settings', $storage_settings);
          $form_state->setRebuild(TRUE);
          $rebuild = TRUE;
          break;

        case 'remove_storage':
          $client_number = $trigger['#attributes']['data-remove'] ?? NULL;
          if (isset($client_number)) {
            $storage_count = $form_state->get('storage_count') - 1;
            $form_state->set('storage_count', $storage_count);
            // Reorder storage client configs.
            $storage_settings = $form_state->get('storage_settings');
            // Note: here, $storage_count = count($storage_settings) - 1.
            $user_input = $form_state->getUserInput();
            for ($config_number = $client_number; $config_number < $storage_count; ++$config_number) {
              $storage_settings[$config_number] = $storage_settings[$config_number + 1];
              // Shift form values.
              $form_state->setValueForElement(
                ['storages_tab', 'storages', $config_number],
                $form_state->getValue(
                  ['storages_tab', 'storages', $config_number + 1]
                )
              );
              $user_input['storages_tab']['storages'][$config_number] = $user_input['storages_tab']['storages'][$config_number + 1];
            }
            unset($storage_settings[$storage_count]);
            $form_state->set('storage_settings', $storage_settings);
            unset($user_input['storages_tab']['storages'][$storage_count]);
            $form_state->setUserInput($user_input);
            $form_state->setRebuild(TRUE);
            $rebuild = TRUE;
          }
          break;

        default:
      }
    }

    // Storage client selection change.
    $client_number = $trigger['#attributes']['data-client'] ?? NULL;
    if (isset($client_number)) {
      $storage_settings = $form_state->get('storage_settings');
      $client_id = $form_state->getValue(
        ['storages_tab', 'storages', $client_number, 'id']
      );
      $client_config = [];
      // Try to get existing settings from config if same id.
      if (!$external_entity_type->isNew()
        && ($external_entity_type->getStorageClientId($client_number) == $client_id)
      ) {
        $client_config =
          $external_entity_type->getStorageClientConfig($client_number);
      }
      $storage_settings[$client_number] = [
        'id' => $client_id,
        'config' => $client_config,
      ];
      $form_state->set('storage_settings', $storage_settings);
      $form_state->setRebuild(TRUE);
      $rebuild = TRUE;
    }

    parent::validateForm($form, $form_state);

    // Validate field mapper settings.
    $field_mapper_id = $form_state->getValue(['field_mapper', 'field_mapper_id']);
    if ($field_mapper_id) {
      $field_mapper = $this->fieldMapperManager->createInstance(
        $field_mapper_id,
        [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $external_entity_type]
      );
      if ($field_mapper instanceof PluginFormInterface) {
        $field_mapper_form_state = SubformState::createForSubform($form['field_mapper']['field_mapper_config'], $form, $form_state);
        $field_mapper->validateConfigurationForm($form['field_mapper']['field_mapper_config'], $field_mapper_form_state);
      }
    }

    // Validate all storage client settings.
    $data_aggregator_id = $form_state->getValue(
      ['storages_tab', 'storages', 'data_aggregator_id'],
      ExternalEntityType::DEFAULT_DATA_AGGREGATOR
    );
    $data_aggregator = $this->dataAggregatorManager->createInstance(
      $data_aggregator_id,
      [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $this->getEntity()]
    );
    if ($data_aggregator instanceof DataAggregatorInterface) {
      $data_aggregator_form_state = SubformState::createForSubform($form['storages_tab']['storages']['aggr'], $form, $form_state);
      $data_aggregator->validateGlobalAggregationForm($form['storages_tab']['storages']['aggr'], $data_aggregator_form_state);
    }

    $storage_count = $form_state->get('storage_count');
    $group_prefixes = [];
    for ($i = 0; $i < $storage_count; ++$i) {
      $storage_client_id = $form_state->getValue(
        ['storages_tab', 'storages', $i, 'id']
      );
      if (!empty($storage_client_id)) {
        // Validate storage client settings.
        $storage_client = $this->storageClientManager->createInstance(
          $storage_client_id
        );
        if ($storage_client instanceof PluginFormInterface) {
          $storage_client_form_state = SubformState::createForSubform(
            $form['storages_tab']['storages'][$i]['config'],
            $form,
            $form_state
          );
          $storage_client->validateConfigurationForm(
            $form['storages_tab']['storages'][$i]['config'],
            $storage_client_form_state
          );
        }
        // Validate storage client aggregation settings.
        if ($data_aggregator instanceof DataAggregatorInterface) {
          $data_aggregator_form_state = SubformState::createForSubform($form['storages_tab']['storages'][$i]['aggr'], $form, $form_state);
          $data_aggregator->validateStorageClientAggregationForm($form['storages_tab']['storages'][$i]['aggr'], $data_aggregator_form_state, $i);
        }
      }
    }

    // If rebuild needed, ignore validation.
    if ($rebuild) {
      $form_state->clearErrors();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $external_entity_type = $this->getEntity();
    // Submit and save field mapper.
    $field_mapper_id = $form_state->getValue(['field_mapper', 'field_mapper_id']);
    if ($field_mapper_id) {
      $field_mapper = $this->fieldMapperManager->createInstance(
        $field_mapper_id,
        [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $this->getEntity()]
      );
      if ($field_mapper instanceof PluginFormInterface) {
        $field_mapper_form_state = SubformState::createForSubform($form['field_mapper']['field_mapper_config'], $form, $form_state);
        $field_mapper->submitConfigurationForm($form['field_mapper']['field_mapper_config'], $field_mapper_form_state);
        $field_mappers = [
          [
            'id' => $field_mapper_id,
            'config' => $field_mapper->getConfiguration(),
          ],
        ];
      }
    }
    $form_state->setValue('field_mappers', $field_mappers);

    // Submit and save all storage clients and aggregation settings.
    $aggr_settings = [
      'global_config' => [],
      'storage_client_config' => [
        [],
      ],
    ];
    $data_aggregator_id = $form_state->getValue(
      ['storages_tab', 'storages', 'data_aggregator_id'],
      ExternalEntityType::DEFAULT_DATA_AGGREGATOR
    );
    $data_aggregator = $this->dataAggregatorManager->createInstance(
      $data_aggregator_id,
      [ExternalEntityTypeInterface::XNTT_TYPE_PROP => $this->getEntity()]
    );
    if ($data_aggregator instanceof DataAggregatorInterface) {
      $data_aggregator_form_state = SubformState::createForSubform($form['storages_tab']['storages']['aggr'], $form, $form_state);
      $data_aggregator->submitGlobalAggregationForm($form['storages_tab']['storages']['aggr'], $data_aggregator_form_state);
    }
    $storage_count = $form_state->get('storage_count');
    $storage_clients = [];
    for ($i = 0; $i < $storage_count; ++$i) {
      $storage_client_id = $form_state->getValue(
        ['storages_tab', 'storages', $i, 'id']
      );
      if (!empty($storage_client_id)) {
        $storage_client = $this->storageClientManager->createInstance(
          $storage_client_id
        );
        if ($storage_client instanceof PluginFormInterface) {
          $storage_client_form_state = SubformState::createForSubform(
            $form['storages_tab']['storages'][$i]['config'],
            $form,
            $form_state
          );
          $storage_client->submitConfigurationForm(
            $form['storages_tab']['storages'][$i]['config'],
            $storage_client_form_state
          );
          $storage_clients[] = [
            'id' => $storage_client_id,
            'config' => $storage_client->getConfiguration(),
          ];
        }
        if ($data_aggregator instanceof DataAggregatorInterface) {
          $data_aggregator_form_state = SubformState::createForSubform($form['storages_tab']['storages'][$i]['aggr'], $form, $form_state);
          $data_aggregator->submitStorageClientAggregationForm($form['storages_tab']['storages'][$i]['aggr'], $data_aggregator_form_state, $i);
        }
      }
    }
    if ($data_aggregator instanceof DataAggregatorInterface) {
      $aggr_settings = $data_aggregator->getConfiguration();
    }
    $form_state->setValue('storage_clients', $storage_clients);
    $form_state->setValue(
      'data_aggregator',
      [
        'id' => $data_aggregator_id,
        'config' => $aggr_settings,
      ]
    );

    // Save other parameters.
    $form_state->setValue(
      'persistent_cache_max_age',
      (int) $form_state->getValue(['caching', 'persistent_cache_max_age'],
      ExternalEntityType::CACHE_DISABLED)
    );
    $form_state->setValue('annotation_entity_type_id', $form_state->getValue([
      'annotations',
      'annotation_settings',
      'annotation_entity_type_id',
    ], NULL));
    $form_state->setValue('annotation_bundle_id', $form_state->getValue([
      'annotations',
      'annotation_settings',
      'annotation_bundle_id',
    ], NULL));
    $form_state->setValue('annotation_field_name', $form_state->getValue([
      'annotations',
      'annotation_settings',
      'annotation_field_name',
    ], NULL));
    $form_state->setValue('inherits_annotation_fields', $form_state->getValue([
      'annotations',
      'annotation_settings',
      'inherits_annotation_fields',
    ], FALSE));

    parent::submitForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $external_entity_type = $this->getEntity();
    $status = $external_entity_type->save();

    if ($status) {
      // Check if mapping is set.
      $fm_config = $external_entity_type->getFieldMapperConfig();
      // Only redirect if ID mapping is set with regular save button.
      if (!empty($fm_config['field_mappings']['id']['value']['mapping'])) {
        $this->messenger->addMessage($this->t('Saved the %label external entity type.', [
          '%label' => $external_entity_type->label(),
        ]));
        if (($trigger = $form_state->getTriggeringElement())
          && ($trigger['#name'] == 'save_and_edit')
        ) {
          $form_state->disableRedirect();
        }
        else {
          $form_state->setRedirect('entity.external_entity_type.collection');
        }
      }
      else {
        $this->messenger->addMessage($this->t('Saved the %label external entity type. Please now set a mapping for the required fields such as the entity identifier field.', [
          '%label' => $external_entity_type->label(),
        ]));
        // Redirect to external entity type edition page.
        $form_state->setRedirect(
          'entity.external_entity_type.edit_form',
          ['external_entity_type' => $external_entity_type->getDerivedEntityTypeId()],
          ['fragment' => 'edit-field-mapper-field-mapper-config-field-mappings-id-value']
        );
      }
    }
    else {
      $this->messenger->addMessage($this->t('The %label external entity type was not saved.', [
        '%label' => $external_entity_type->label(),
      ]));
    }
  }

  /**
   * Checks if the entity type already exists.
   *
   * @param string $entity_type_id
   *   The entity type id to check.
   *
   * @return bool
   *   TRUE if already exists, FALSE otherwise.
   */
  public function exists($entity_type_id) {
    if ($this->entityTypeManager->getDefinition($entity_type_id, FALSE)) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Gets the form entity.
   *
   * The form entity which has been used for populating form element defaults.
   *
   * @return \Drupal\external_entities\Entity\ExternalEntityTypeInterface
   *   The current form entity.
   */
  public function getEntity() {
    /** @var \Drupal\external_entities\Entity\ExternalEntityTypeInterface $entity */
    $entity = $this->entity;
    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityFromRouteMatch(RouteMatchInterface $route_match, $entity_type_id) {
    if ($route_match->getParameter($entity_type_id) !== NULL) {
      $entity_id = $route_match->getParameter($entity_type_id);
      $entity = $this->entityTypeManager->getStorage($entity_type_id)->load($entity_id);
      if ($entity) {
        return $entity;
      }
    }

    return parent::getEntityFromRouteMatch($route_match, $entity_type_id);
  }

  /**
   * Gets the annotation entity type id options.
   *
   * @return array
   *   Associative array of entity type labels, keyed by their ids.
   */
  public function getAnnotationEntityTypeIdOptions() {
    $options = [];

    $definitions = $this->entityTypeManager->getDefinitions();
    /** @var \Drupal\Core\Entity\EntityTypeInterface $definition */
    foreach ($definitions as $entity_type_id => $definition) {
      if ($entity_type_id !== 'external_entity' && $definition instanceof ContentEntityType) {
        $options[$entity_type_id] = $definition->getLabel();
      }
    }

    return $options;
  }

  /**
   * Gets the annotation bundle options.
   *
   * @param string $entity_type_id
   *   (optional) The entity type id.
   *
   * @return array
   *   Associative array of bundle labels, keyed by their ids.
   */
  public function getAnnotationBundleIdOptions($entity_type_id = NULL) {
    $options = [];

    if ($entity_type_id) {
      $bundles = $this->entityTypeBundleInfo->getBundleInfo($entity_type_id);
      foreach ($bundles as $bundle_id => $bundle) {
        $options[$bundle_id] = $bundle['label'];
      }
    }

    return $options;
  }

  /**
   * Gets the annotation field options.
   *
   * @param string $entity_type_id
   *   The entity type id.
   * @param string $bundle_id
   *   The bundle id.
   *
   * @return array
   *   Associative array of field labels, keyed by their ids.
   */
  public function getAnnotationFieldOptions($entity_type_id, $bundle_id) {
    $options = [];

    if ($entity_type_id && $bundle_id) {
      $fields = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle_id);
      foreach ($fields as $id => $field) {
        if ($field->getType() === 'entity_reference' && $field->getSetting('target_type') === $this->getEntity()->getDerivedEntityTypeId()) {
          $options[$id] = $field->getLabel() . ' (' . $field->getName() . ')';
        }
      }
    }

    return $options;
  }

}
