<?php

namespace Drupal\external_entities\Plugin\ExternalEntities\DataAggregator;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\external_entities\DataAggregator\DataAggregatorBase;
use Drupal\external_entities\DataAggregator\DataAggregatorInterface;
use Drupal\external_entities\Entity\ExternalEntityInterface;
use Drupal\external_entities\Entity\ExternalEntityTypeInterface;
use Drupal\external_entities\FieldMapper\FieldMapperInterface;
use Drupal\external_entities\Plugin\PluginFormTrait;
use JsonPath\InvalidJsonException;
use JsonPath\InvalidJsonPathException;
use JsonPath\JsonObject;

/**
 * External entities data aggregator by groups.
 *
 * @DataAggregator(
 *   id = "group",
 *   label = @Translation("Group Aggregator"),
 *   description = @Translation("Aggregate data from grouped data sources.")
 * )
 */
class GroupAggregator extends DataAggregatorBase implements DataAggregatorInterface {

  use PluginFormTrait;
  use MessengerTrait;

  /**
   * Remapped storage client instances.
   *
   * @var \Drupal\external_entities\StorageClient\StorageClientInterface[]
   */
  protected $remappedStorageClients = [];

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'global_config' => [],
      'storage_client_config' => [
        [
          'groups' => [],
          'group_prefix_strip' => FALSE,
          'id_field' => '',
          'merge' => 'keep',
          'merge_as_member' => '',
          'merge_join' => '',
          'readonly' => FALSE,
          'client_number' => 0,
        ],
      ],
    ];
  }

  /**
   * Returns current external entity type remapped storage clients.
   *
   * A remapped storage client is storage client from which the associated
   * external entity type has been replaced by a virtual one (ie. living in
   * memory) that has no other storage client that the remapped one and which
   * id field mapping has been adjusted to meet aggregation settings.
   *
   * @return \Drupal\external_entities\StorageClient\StorageClientInterface[]
   *   An array of remapped storage clients matching current external entity
   *   storage clients. If no current external entity is set, returns an empty
   *   array.
   */
  protected function getRemappedStorageClients() :array {
    if (empty($this->externalEntityType)) {
      // Warn for invalid initialization.
      $this->logger->warning(
        'GroupAggregator::getRemappedStorageClients() called without proper initialization of the external entity type member'
      );
      return [];
    }

    if (!$this->remappedStorageClients) {
      // Get first client.
      $this->remappedStorageClients = [
        $this->externalEntityType->getStorageClient(0),
      ];
      $mapper_id = $this->externalEntityType->getFieldMapperId(0);
      $mapper_config = $this->externalEntityType->getFieldMapperConfig(0);
      // $mapper_config = [];
      $storage_count = $this->externalEntityType->getStorageClientCount();
      for ($i = 1; $i < $storage_count; ++$i) {
        $aggr_settings = $this->getConfiguration()['storage_client_config'][$i];
        // Create a specific virtual external entity type for each storage
        // client to handle join fields in field mappings.
        // Indeed, when we use join fields, we only need one storage client and
        // we may need to map id to a different client source field.
        $virtual_xntt_type = $this->externalEntityType->createDuplicate();
        // Remove field mappers.
        while ($c = $virtual_xntt_type->getFieldMapperCount()) {
          $virtual_xntt_type->removeFieldMapperId($c - 1);
        }
        // Remove storage clients.
        while ($c = $virtual_xntt_type->getStorageClientCount()) {
          $virtual_xntt_type->removeStorageClientId($c - 1);
        }
        // @todo Also clear aggregation settings.
        // Put back original id to have an external entity type name when
        // loading external entity instances.
        $virtual_xntt_type->set('id', $this->externalEntityType->id());
        // Create a new instance of field mapper.
        // @todo Maybe we only need to map id field? To investigate.
        // Note: for field mapper config,
        // ExternalEntityTypeInterface::XNTT_TYPE_PROP
        // is also set by ExternalEntityType::setFieldMapperConfig() but we set
        // it here in case the code after does not call setFieldMapperConfig()
        // anymore.
        $mapper_config[ExternalEntityTypeInterface::XNTT_TYPE_PROP] = $virtual_xntt_type;
        $field_mapper = $this->fieldMapperManager->createInstance(
          $mapper_id,
          $mapper_config
        );
        // Remap ID if needed.
        if (!empty($aggr_settings['id_field'])) {
          $field_mapper->setFieldMapping(
            'id',
            'value',
            $aggr_settings['id_field'],
            FieldMapperInterface::MAPPING_DIRECT
          );
        }
        $virtual_xntt_type->addFieldMapperId('simple');
        $virtual_xntt_type->setFieldMapperConfig(
          $field_mapper->getConfiguration()
        );
        $storage_client_id = $this->externalEntityType->getStorageClientId($i);
        $storage_client_config = $this->externalEntityType->getStorageClientConfig($i);
        $storage_client_config[ExternalEntityTypeInterface::XNTT_TYPE_PROP] = $virtual_xntt_type;
        $this->remappedStorageClients[$i] = $this->storageClientManager->createInstance(
          $storage_client_id,
          $storage_client_config
        );
      }
    }
    return $this->remappedStorageClients;
  }

  /**
   * Returns the storage clients matching an entity id for aggregation.
   *
   * Only return remapped storage clients that are relevant to the provided
   * entity identifier. Since some clients may use alternative fields as
   * joins, they will not be filtered here but they should be fitered
   * afterward, from the related join field values.
   * Returned storage clients are still indexed with their initial index used
   * in the external entity type.
   *
   * @param mixed $id
   *   Identifier value of current external entity considered.
   * @param bool $exclude_read_only
   *   Excludes read-only storage client from the returend list.
   *
   * @return \Drupal\external_entities\StorageClient\StorageClientInterface[]
   *   An array of remapped storage clients indexed using their original
   *   external entity type storage client index. Only storage clients relevants
   *   to the given identifier are returned which means there could be expected
   *   index gaps in the returned array as irrelevant storage clients are
   *   filtered out. If no current external entity is set, returns an empty
   *   array.
   */
  protected function filterAggregationStorageClientsForId(
    $id,
    bool $exclude_read_only = FALSE
  ) :array {
    $clients = $this->getRemappedStorageClients();
    $aggr_clients = [];
    foreach ($clients as $client_index => $client) {
      $aggr_settings =
        $this->getConfiguration()['storage_client_config'][$client_index];
      if ($exclude_read_only && !empty($aggr_settings['readonly'])) {
        // Read-only, skip.
        continue 1;
      }
      $groups = $aggr_settings['groups'] ?? [];
      if (empty($groups)) {
        // Not a group-specific client.
        $aggr_clients[$client_index] = $client;
      }
      else {
        // Check if a group prefix macthes this entity ID.
        $re = '/^\Q' . implode('\E|^\Q', $groups) . '\E/';
        if (1 === preg_match($re, $id)) {
          $aggr_clients[$client_index] = $client;
        }
      }
    }
    return $aggr_clients;
  }

  /**
   * {@inheritdoc}
   */
  public function load(string|int $id) :array|null {
    $entities = $this->loadMultiple([$id]);
    return array_key_exists($id, $entities)
      ? $entities[$id]
      : NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function loadMultiple(array $ids = NULL) :array {
    if (empty($this->externalEntityType)) {
      // Warn for invalid initialization.
      $this->logger->warning(
        'GroupAggregator::loadMultiple() called without proper initialization of the external entity type member'
      );
      return [];
    }

    $clients = $this->getRemappedStorageClients();
    $aggr_config = [];
    foreach ($clients as $client_index => $client) {
      $aggr_config[$client_index] =
        $this->getConfiguration()['storage_client_config'][$client_index];
    }

    $no_groups = empty($aggr_config[0]['groups']);

    // Check if we are working with groups and set groups.
    if ($no_groups) {
      // If first storage client has no group, then it is the reference.
      // Only its entities will be taken into account.
      $clients_by_group = ['' => $clients];
    }
    else {
      // Otherwise we work by groups.
      // Get each group reference and set of clients.
      $clients_by_group = [];
      foreach ($clients as $client_index => $client) {
        $groups = $aggr_config[$client_index]['groups'] ?? [];
        if (empty($groups)) {
          // No groups, add client to every current group.
          foreach (array_keys($clients_by_group) as $group) {
            $clients_by_group[$group][$client_index] = $client;
          }
        }
        else {
          // Add client to each of its groups.
          foreach ($groups as $group) {
            $clients_by_group[$group][$client_index] = $client;
          }
        }
      }
    }

    $entities = [];
    foreach ($clients_by_group as $group => $group_clients) {
      $group_len = strlen($group);
      // Get the reference storage client.
      reset($group_clients);
      $ref_client_number = key($group_clients);
      $ref_client = current($group_clients);
      // We will process the reference storage client separately.
      // Unset the ref. client to keep indexes but remove it from the set.
      unset($group_clients[$ref_client_number]);
      $id_field = $this
        ->externalEntityType
        ->getFieldMapper()
        ->getMappedFieldName('id', 'value');

      // Check if the reference client uses group filtering.
      $groups = $aggr_config[$ref_client_number]['groups'] ?? [];
      $valid_ids = $ids;
      if (is_array($valid_ids)) {
        if (!empty($group)) {
          // Filter requested ids for this group.
          $valid_ids = array_filter(
            $valid_ids,
            function ($id) use ($group_len, $group) {
              return (strncmp($id, $group, $group_len) === 0);
            }
          );
        }
        // Check if group prefix needs to be stripped.
        if (!empty($aggr_config[$ref_client_number]['group_prefix_strip'])
          && !empty($group)
        ) {
          // Use virtual pefix: strip.
          $valid_ids = array_map(
            function ($x) use ($group_len) {
              return substr($x, $group_len);
            },
            $valid_ids
          );
        }
      }
      if (!empty($valid_ids) || (NULL === $valid_ids)) {
        $client_entities = $ref_client->loadMultiple($valid_ids) ?? [];
      }
      else {
        $client_entities = [];
      }

      // Filter ids to only loaded ids.
      $valid_ids = array_keys($client_entities);
      if (!empty($aggr_config[$ref_client_number]['group_prefix_strip'])
        && !empty($group)
      ) {
        // Use virtual pefix: unstrip.
        $valid_ids = array_map(
          function ($x) use ($group) {
            return $group . $x;
          },
          $valid_ids
        );
        $remapped_entities = [];
        foreach ($client_entities as $client_entity_id => $client_entity) {
          if (0 < strlen($client_entity_id ?? '')) {
            $client_entity_id = $group . $client_entity_id;
            $remapped_entities[$client_entity_id] = $client_entity;
          }
        }
        $client_entities = $remapped_entities;
      }
      $valid_ids = array_combine($valid_ids, $valid_ids);
      // Merge arrays but keep keys as they are, especially for integer keys as
      // keys are identifiers. We should only get new keys and there should
      // not be any override here.
      foreach ($client_entities as $client_entity_id => $client_entity) {
        $entities[$client_entity_id] = $client_entity;
      }

      // Aggregate values from other storage clients.
      $this->aggregateSubStorageClientsData(
        $entities,
        $group_clients,
        $aggr_config,
        $valid_ids,
        $id_field
      );
    }
    return $entities;
  }

  /**
   * Aggregate data from storage client of a given group to current data.
   *
   * @param array &$entities
   *   Raw reference client entities array keyed by ids that will contain
   *   aggregated data.
   * @param StorageClientInterface[] $group_clients
   *   Array of remapped storage clients of a given group.
   * @param array $aggr_config
   *   Aggregation config containing all storage client aggregation settings.
   * @param array $valid_ids
   *   An array of entity identifiers that correspond to the group.
   * @param string $id_field
   *   The name of the reference storage client id field used for entities.
   */
  protected function aggregateSubStorageClientsData(
    array &$entities,
    array $group_clients,
    array $aggr_config,
    array $valid_ids,
    string $id_field,
  ) :void {
    foreach ($group_clients as $client_index => $client) {
      // First, get the list of "ids" to load from current client.
      // $client_ids keys are reference entity ids, values are array of
      // current client data "ids" to match.
      $client_ids = [];
      foreach ($entities as $raw_entity) {
        if (!empty($aggr_config[$client_index]['merge_join'])) {
          // Merge using a specific field on current data as id.
          $group_join_field = $aggr_config[$client_index]['merge_join'];
          // Check if we got a JSONPath.
          if (preg_match('/^\$[.\[]/', $group_join_field)) {
            $jpath = new JsonObject($raw_entity, TRUE);
            try {
              $join_value = $jpath->get($group_join_field);
              if (FALSE === $join_value) {
                $join_value = [];
              }
              $client_ids[$raw_entity[$id_field]] =
                is_array($join_value)
                ? $join_value
                : [$join_value];
            }
            catch (InvalidJsonException | InvalidJsonPathException $e) {
              // JSONPath mapping failed. Report.
              \Drupal::logger('external_entities')->error(
                'Invalid JSONPath: '
                . $e
                . "\nExternal Entity Type: "
                . $this->externalEntityType->id()
                . "\nStorage client: $client_index\nJSONPath: "
                . print_r($group_join_field, TRUE)
                . "\nData: "
                . print_r($raw_entity, TRUE)
              );
              $client_ids[$raw_entity[$id_field]] = [];
            }
          }
          elseif (isset($raw_entity[$group_join_field])) {
            $client_ids[$raw_entity[$id_field]] =
              is_array($raw_entity[$group_join_field])
              ? $raw_entity[$group_join_field]
              : [$raw_entity[$group_join_field]];
          }
          // Else, it means the raw entity does not have any value for that
          // join field and no value can be added.
        }
        else {
          // Use current entity id as id value for current client.
          $client_ids = array_map(
            function ($vid) {
              return [$vid];
            },
            $valid_ids
          );
        }
      }

      // Get id field used on current client.
      $query_id_field =
        empty($aggr_config[$client_index]['id_field'])
        ? $id_field
        : $aggr_config[$client_index]['id_field'];
      // Get all storage client entitites at once.
      // Note: the given field to filter is a raw field, not a Drupal
      // one. However, the query filter only supports Drupal fields.
      // To get around that problem, the client has been configured
      // with an altered field mapping that maps the Drupal id field
      // to the current client field to use for joining.
      // See self::getRemappedStorageClients() for details.
      $parameters = [
        [
          'field' => 'id',
          'value' => array_merge(...array_values($client_ids)),
          'operator' => 'IN',
        ],
      ];
      // @todo The sort order could be set by config.
      $query_entities = $client->query(
        $parameters,
        [
          ['field' => $query_id_field, 'direction' => 'ASC'],
        ]
      ) ?? [];
      // Here, $query_entities is supposed to contain all entities matching
      // the given "ids" in an array. However, the "id" field used may not
      // provide unique matches: it is possible that more than one returned
      // entity share the same "id". That "id" field can be either
      // $aggr_config[$client_index]['id_field'] if not empty or $id_field
      // otherwise.
      // Now, we will use $client_ids array to match current entity data
      // with the new one. If we got more than one matching new record, we
      // will handle 2 cases:
      // 1) the new records must be added in a subfield, no problem there.
      // 2) the new records must be merged to current one. Each record will
      // be merged one after the other it the order returned by the client
      // using $query_id_field sort order.
      // Sets default merge behavior.
      $aggr_config[$client_index]['merge'] ??= 'keep';
      if ('sub' == $aggr_config[$client_index]['merge']) {
        // Store corresponding client entities as member.
        $member =
          ($aggr_config[$client_index]['merge_as_member'] ?? '')
          ?: $client->getPluginId() . '_' . $client_index;
        foreach ($client_ids as $source_id => $matching_ids) {
          $entities[$source_id][$member] = array_filter(
            $query_entities,
            function ($qe) use ($query_id_field, $matching_ids) {
              return (in_array($qe[$query_id_field], $matching_ids));
            }
          );
        }
      }
      else {
        foreach ($client_ids as $source_id => $matching_ids) {
          // Gather current client entities matching $source_id.
          $matching_entities = array_filter(
            $query_entities,
            function ($qe) use ($query_id_field, $matching_ids) {
              return (in_array($qe[$query_id_field], $matching_ids));
            }
          );
          if (!empty($matching_entities)) {
            // We got something to merge, check for override.
            if ('keep' == $aggr_config[$client_index]['merge']) {
              // No override of fields already set.
              foreach ($matching_entities as $matching_entity) {
                $entities[$source_id] = static::mergeArrays(
                  $matching_entity,
                  $entities[$source_id],
                  TRUE,
                  FALSE,
                  FALSE,
                  FALSE
                );
              }
            }
            elseif ('over' == $aggr_config[$client_index]['merge']) {
              foreach ($matching_entities as $matching_entity) {
                // Override all except id.
                unset($matching_entity[$id_field]);
                $entities[$source_id] = static::mergeArrays(
                  $entities[$source_id],
                  $matching_entity,
                  TRUE,
                  FALSE,
                  FALSE,
                  FALSE
                );
              }
            }
            elseif ('oven' == $aggr_config[$client_index]['merge']) {
              // Override only existing fields except id.
              foreach ($matching_entities as $matching_entity) {
                unset($matching_entity[$id_field]);
                $entities[$source_id] = static::mergeArrays(
                  $entities[$source_id],
                  $matching_entity,
                  TRUE,
                  FALSE,
                  FALSE,
                  TRUE
                );
              }
            }
            elseif ('ovem' == $aggr_config[$client_index]['merge']) {
              // Override empty fields only.
              foreach ($matching_entities as $matching_entity) {
                $entities[$source_id] = static::mergeArrays(
                  $entities[$source_id],
                  $matching_entity,
                  TRUE,
                  TRUE,
                  FALSE,
                  FALSE
                );
              }
            }
            else {
              // Unknown merge method, warn.
              \Drupal::logger('external_entities')->warning(
                'Unsupported merge method "'
                . $aggr_config[$client_index]['merge']
                . '" for client '
                . $client_index
                . ' of external entity type '
                . $this->externalEntityType->id()
                . '.'
              );
            }
          }
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function save(ExternalEntityInterface $entity) :int {
    $return_value = 0;
    $aggr_config = $this->getConfiguration()['storage_client_config'];
    // Filter aggregation plugins.
    $aggr_plugins = $this->filterAggregationStorageClientsForId($entity->id(), TRUE);

    $original_id = $entity->id();
    foreach ($aggr_plugins as $plugin_index => $plugin) {
      // Set new id to the orignal one to later detect changes.
      $new_id = $original_id;
      // Check if identifier should be adjusted (data joins).
      if (!empty($aggr_config[$plugin_index]['id_field'])) {
        // Using a join field instead of the default identifier field.
        $raw_data = $entity->toRawData();
        $id_field = $aggr_config[$plugin_index]['id_field'];
        $new_id = $raw_data[$id_field] ?? NULL;
      }
      elseif (!empty($aggr_config[$plugin_index]['group_prefix_strip'])) {
        // Default identifier should be stipped.
        $groups = $aggr_config[$plugin_index]['groups'] ?? [];
        // Remove group prefixes.
        $re = '/^\Q' . implode('\E|^\Q', $groups) . '\E/';
        $new_id = preg_replace($re, '', $original_id);
      }

      if (!isset($new_id) || ('' == $new_id)) {
        // The join identifier is not available, skip.
        continue 1;
      }
      if ($new_id === $original_id) {
        // No id change.
        $return_value = max($plugin->save($entity), $return_value);
      }
      else {
        // Id change.
        // Save using the given id.
        $entity->id = $new_id;
        $entity->setOriginalId($new_id);
        $return_value = max($plugin->save($entity), $return_value);
        // Put back previous value.
        $entity->id = $original_id;
        $entity->setOriginalId($original_id);
      }
    }

    return $return_value;
  }

  /**
   * {@inheritdoc}
   */
  public function delete(ExternalEntityInterface $entity) {
    // Filter aggregation clients.
    $aggr_clients = $this->filterAggregationStorageClientsForId($entity->id(), TRUE);

    $original_id = $entity->id();
    foreach ($aggr_clients as $client_index => $client) {
      $aggr_settings = $this
        ->getExternalEntityType()
        ->getStorageClientAggregation($client_index);
      // Check if it is a member object: no edit.
      if (!empty($aggr_settings['merge'])
          && ('sub' == $aggr_settings['merge'])
      ) {
        continue;
      }

      // Set $new_id to the orignal one to later detect changes.
      $new_id = $original_id;
      // Check if identifier should be adjusted.
      if (!empty($aggr_settings['id_field'])) {
        // Using a join field instead of the default identifier field.
        $raw_data = $entity->toRawData();
        $id_field = $aggr_settings['id_field'];
        $new_id = $raw_data[$id_field] ?? NULL;
      }
      elseif (!empty($aggr_settings['group_prefix_strip'])) {
        // Default identifier should be stipped.
        $groups = $aggr_settings['groups'] ?? [];
        // Remove group prefixes.
        $re = '/^\Q' . implode('\E|^\Q', $groups) . '\E/';
        $new_id = preg_replace($re, '', $original_id);
      }

      if (!isset($new_id) || ('' == $new_id)) {
        // The join identifier is not available, skip.
        continue 1;
      }
      if ($new_id === $original_id) {
        // No id change.
        $client->delete($entity);
      }
      else {
        // Id change.
        // Delete using the given id.
        $entity->id = $new_id;
        $entity->setOriginalId($new_id);
        $client->delete($entity);
        // Put back previous value.
        $entity->id = $original_id;
        $entity->setOriginalId($original_id);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function query(
    array $parameters = [],
    array $sorts = [],
    ?int $start = NULL,
    ?int $length = NULL
  ) :array {
    if (empty($this->externalEntityType)) {
      // Warn for invalid initialization.
      $this->logger->warning(
        'GroupAggregator::loadMultiple() called without proper initialization of the external entity type member'
      );
      return [];
    }

    $clients = $this->getRemappedStorageClients();
    $aggr_config = [];
    foreach ($clients as $client_index => $client) {
      $aggr_config[$client_index] =
        $this->getConfiguration()['storage_client_config'][$client_index];
    }

    $no_groups = empty($aggr_config[0]['groups']);

    // Check if we are working with groups and set groups.
    if ($no_groups) {
      // If first storage client has no group, then it is the reference.
      // Only its entities will be taken into account.
      $clients_by_group = ['' => $clients];
    }
    else {
      // Otherwise we work by groups.
      // Get each group reference and set of clients.
      $clients_by_group = [];
      foreach ($clients as $client_index => $client) {
        $groups = $aggr_config[$client_index]['groups'] ?? [];
        if (empty($groups)) {
          // No groups, add client to every current group.
          foreach (array_keys($clients_by_group) as $group) {
            $clients_by_group[$group][$client_index] = $client;
          }
        }
        else {
          // Add client to each of its groups.
          foreach ($groups as $group) {
            $clients_by_group[$group][$client_index] = $client;
          }
        }
      }
    }

    $entities = [];
    foreach ($clients_by_group as $group => $group_clients) {
      $group_len = strlen($group);
      // Get the reference storage client.
      reset($group_clients);
      $ref_client_number = key($group_clients);
      $ref_client = current($group_clients);
      // We will process the reference storage client separately.
      // Unset the ref. client to keep indexes but remove it from the set.
      unset($group_clients[$ref_client_number]);
      $id_field = $this
        ->externalEntityType
        ->getFieldMapper()
        ->getMappedFieldName('id', 'value');

      // Check if the reference client uses group filtering.
      $groups = $aggr_config[$ref_client_number]['groups'] ?? [];
      // Query ref client without limitations because they can not be applied
      // before aggregating all reference sources.
      $client_entities = $ref_client->query($parameters);

      // Get current client id field.
      $client_id_field =
        $aggr_config[$ref_client_number]['id_field']
        ?: $id_field;

      // Filter ids by group and remap result by ids.
      $remapped_entities = [];
      $valid_ids = [];
      $group_len = strlen($group);
      foreach ($client_entities as $client_entity) {
        $id = $client_entity[$client_id_field];
        if ($group_len) {
          if (strncmp($id, $group, $group_len) === 0) {
            $valid_ids[] = $id;
            $remapped_entities[$id] = $client_entity;
          }
        }
        else {
          $valid_ids[] = $id;
          $remapped_entities[$id] = $client_entity;
        }
      }
      $client_entities = $remapped_entities;

      // Ajust id filter according to prefix settings.
      if (!empty($aggr_config[$ref_client_number]['group_prefix_strip'])
        && !empty($group)
      ) {
        // Use virtual pefix: unstrip.
        $valid_ids = array_map(
          function ($x) use ($group) {
            return $group . $x;
          },
          $valid_ids
        );
        $remapped_entities = [];
        foreach ($client_entities as $client_entity) {
          if (0 < strlen($client_entity[$id_field] ?? '')) {
            $client_entity[$id_field] = $group . $client_entity[$id_field];
            $remapped_entities[$client_entity[$id_field]] = $client_entity;
          }
        }
        $client_entities = $remapped_entities;
      }
      $valid_ids = array_combine($valid_ids, $valid_ids);
      // Merge arrays but keep keys as they are, especially for integer keys as
      // keys are identifiers. We should only get new keys and there should
      // not be any override here.
      foreach ($client_entities as $client_entity_id => $client_entity) {
        $entities[$client_entity_id] = $client_entity;
      }

      // Aggregate values from other storage clients.
      $this->aggregateSubStorageClientsData(
        $entities,
        $group_clients,
        $aggr_config,
        $valid_ids,
        $id_field
      );
    }

    // Process sorting.
    if (!empty($sorts)) {
      // Get list of fields involved in sorting.
      $field_names = [];
      foreach ($sorts as $sort) {
        if (!is_array($sort) || empty($sort['field'])) {
          // Invalid sort parameter.
          continue;
        }
        $field_name = strstr($sort['field'], '.', TRUE);
        if (FALSE === $field_name) {
          $field_name = $sort['field'];
        }
        $field_names[] = $field_name;
      }
    }
    // Continue processing sorts if needed.
    if (!empty($field_names)) {
      // Extract Drupal raw entity data for sorting.
      $drupal_raw_entities = [];
      $field_mapper = $this->externalEntityType->getFieldMapper();
      foreach ($entities as $id => $raw_entity) {
        $drupal_raw_entities[] = $field_mapper->extractEntityValuesFromRawData(
          $raw_entity,
          $field_names
        );
      }
      // Sort.
      storageClientBase::sortEntities($drupal_raw_entities, $sorts);
      $remapped_entities = [];
      foreach ($drupal_raw_entities as $drupal_raw_entity) {
        $remapped_entities[] = $entities[$drupal_raw_entity['id']['value']];
      }
      $entities = $remapped_entities;
    }
    else {
      // Clear id indexation.
      $entities = array_values($entities);
    }

    // Process paging.
    if (isset($length)) {
      $entities = array_splice($entities, $start ?? 0, $length);
    }
    elseif (isset($start)) {
      $entities = array_splice($entities, $start);
    }

    return $entities;
  }

  /**
   * {@inheritdoc}
   */
  public function countQuery(array $parameters = []) :int {
    $count = 0;

    $plugins = $this->getRemappedStorageClients();
    $config = $this->getConfiguration();
    $aggr_config = $config['storage_client_config'];

    // Count entities from plugins.
    $count = 0;
    if (empty($aggr_config[0]['groups'])) {
      // If first storage client has no group, then it is the reference.
      // Only its entity identifiers will be taken into account.
      $count += $plugins[0]->countQuery($parameters);
    }
    else {
      // Otherwise we work by groups.
      // Get each group reference and count.
      $counted_groups = [];
      foreach ($plugins as $plugin_index => $plugin) {
        $groups = $aggr_config[$plugin_index]['groups'] ?? [];
        if (empty($groups)) {
          // No groups, skip.
          continue;
        }
        else {
          // Count each new group.
          foreach ($groups as $group) {
            if (empty($counted_groups[$group])) {
              // @todo We can't check if the counted entity identifiers match
              // their group. Maybe we should add an id filter?
              $count += $plugin->countQuery($parameters);
              $counted_groups[$group] = TRUE;
            }
          }
        }
      }
    }

    return $count;
  }

  /**
   * {@inheritdoc}
   */
  public function buildStorageClientAggregationForm(array $form, FormStateInterface $form_state, int $client_index) {
    $aggr_selector = uniqid('aggr', TRUE);
    $aggr_config = $this->getConfiguration();
    // Data aggregation settings.
    $form['groups'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Group prefix(es)'),
      '#description' => $this->t('Optional: you can set a group prefix for filtering or group aggregation. Multiple prefixes can be specified separated by semicolons (;).'),
      '#default_value' => implode(';', $aggr_config['storage_client_config'][$client_index]['groups'] ?? []),
      '#attributes' => [
        'data-aggr-selector' => $aggr_selector . '-groups',
      ],
    ];
    $form['group_prefix_strip'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Virtual group prefix'),
      '#description' => $this->t('"Virtual group prefix" means the group prefix is not present on the storage client side but only visible from the External Entity side.'),
      '#default_value' => $aggr_config['storage_client_config'][$client_index]['group_prefix_strip'] ?? FALSE,
      '#states' => [
        'invisible' => [
          ':input[data-aggr-selector="' . $aggr_selector . '-groups"]' => ['value' => ''],
        ],
      ],
    ];

    // The use of override or other identifier field can not work with
    // the first storage client as no field has been loaded for it.
    if ($client_index) {
      $form['merge_join'] = [
        '#type' => 'textfield',
        '#title' => $this->t(
          'Previous storage clients field name that provides the identifier value for this storage client (join field)'
        ),
        '#description' => $this->t(
          'Leave empty to use the default identifier. Otherwise, any (text/numeric) field provided by previous storage clients can be used here as identifier value for this storage client as well as a JSONPath expression.'
        ),
        '#default_value' => $aggr_config['storage_client_config'][$client_index]['merge_join'] ?? NULL,
      ];

      $form['id_field'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Storage source (id) field to use to join data'),
        '#description' => $this->t(
          'Source field name that will hold the value used to join data with the given (or default) previous storage identifier field.'
        ),
        '#default_value' => $aggr_config['storage_client_config'][$client_index]['id_field'] ?? '',
      ];

      $form['merge'] = [
        '#type' => 'select',
        '#title' => $this->t('How to merge client data'),
        '#options' => [
          'keep' => $this->t('Keep existing field values (no override)'),
          'over' => $this->t('Override previous field values'),
          'ovem' => $this->t('Override previous field values if empty'),
          'sub'  => $this->t('As a sub-object'),
        ],
        '#description' => $this->t(
          'If set, entity field values provided by this storage client will override existing values with the same field name provided by previous storage clients (except for the identifier and join fields).'
        ),
        '#default_value' => $aggr_config['storage_client_config'][$client_index]['merge'] ?? 'keep',
        '#attributes' => [
          'data-aggr-selector' => $aggr_selector . '-merge',
        ],
      ];

      $form['merge_as_member'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Field name to store sub-object(s) in an array'),
        '#description' => $this->t(
          'Machine name of the field that will hold the array of sub-objects.'
        ),
        '#default_value' => $aggr_config['storage_client_config'][$client_index]['merge_as_member'] ?? '',
        '#states' => [
          'visible' => [
            ':input[data-aggr-selector="' . $aggr_selector . '-merge"]' => ['value' => 'sub'],
          ],
        ],
      ];
    }
    else {
      $form['id_field'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Storage (source) field name to use as identifier for this storage client'),
        '#description' => $this->t(
          'This field would be used for group prefix filtering and should correspond to the field mapped to the external entity id field. If not set, it will try to use the id mapping unless it is not a direct mapping.'
        ),
        '#default_value' => $aggr_config['storage_client_config'][$client_index]['id_field'] ?? NULL,
      ];
    }

    $form['readonly'] = [
      '#type' => 'checkbox',
      '#title' => $this->t(
        'Read only (no create/update/delete performed on storage client)'
      ),
      '#default_value' => $aggr_config['storage_client_config'][$client_index]['readonly'] ?? FALSE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateStorageClientAggregationForm(
    array &$form,
    FormStateInterface $form_state,
    int $client_index
  ) {
    $aggr_settings = $form_state->getValues();
    $aggr_settings['groups'] = array_filter(
      preg_split('/\s*;\s*/', trim($aggr_settings['groups'] ?? '')),
      'strlen'
    );
    $form_state->setValues($aggr_settings);

    // Check prefixes...
    // Make sure a given prefix does not appear for the first time in a set.
    $group_prefixes = [];
    $new_prefixes = [];
    foreach ($aggr_settings['groups'] as $prefix) {
      if (!array_key_exists($prefix, $group_prefixes)) {
        $new_prefixes[$prefix] = $prefix;
      }
    }
    if ((!empty($new_prefixes))
        && (count($new_prefixes) != count($aggr_settings['groups']))
    ) {
      $form_state->setErrorByName(
        'groups',
        $this->t(
          'The group prefix(es) %prefix appear(s) for the first time in a set of group prefixes where at least one appeared before. However, when multiple group prefixes are specified, they should ALL appear only for the first time or have appeared at least once before (no mix between new and already defined prefixes).',
          [
            '%prefix' => implode(
              ', ',
              array_diff_key($new_prefixes, $group_prefixes)
            ),
          ]
        )
      );
    }
    $group_prefixes += $new_prefixes;

    // Make sure no prefix is part of another one.
    foreach ($new_prefixes as $group_prefix1) {
      foreach ($group_prefixes as $group_prefix2) {
        if (($group_prefix1 != $group_prefix2)) {
          if (strncmp($group_prefix2, $group_prefix1, strlen($group_prefix1)) === 0) {
            $form_state->setErrorByName(
              'groups',
              $this->t(
                'In external entity storage client "Multiple storages", the group prefix %prefix1 is also a prefix for group prefix %prefix2. Prefixes must be strictly distinct.',
                [
                  '%prefix1' => $group_prefix1,
                  '%prefix2' => $group_prefix2,
                ]
              )
            );
          }
          elseif (strncmp($group_prefix1, $group_prefix2, strlen($group_prefix2)) === 0) {
            $form_state->setErrorByName(
              'groups',
              $this->t(
                'In external entity storage client "Multiple storages", the group prefix %prefix1 is also a prefix for group prefix %prefix2. Prefixes must be strictly distinct.',
                [
                  '%prefix1' => $group_prefix2,
                  '%prefix2' => $group_prefix1,
                ]
              )
            );
          }
        }
      }
    }
  }

}
