<?php

namespace Drupal\external_entities\Plugin;

use Drupal\Component\Plugin\ConfigurableInterface;
use Drupal\Core\Form\FormStateInterface;

/**
 * Provides default implementations for plugin configuration forms.
 *
 * @see \Drupal\Core\Plugin\PluginFormInterface
 */
trait PluginFormTrait {

  /**
   * Handles AJAX subform update.
   *
   * When using AJAX callbacks in subform, the returned $form object is not
   * the original subform but the full form. Therefore, it is not possible to
   * simply use the knowledge of the subform structure to find out what part of
   * $form should be returned.
   *
   * The trick here is to use the triggering element that comes from the subform
   * and climb back up in the form tree until we find the wrapper element to
   * update and return it.
   *
   * Usage:
   * @code
   *   $form['maincntr']['subcontainer']['#attributes']['id'] = uniqid('xntt');
   *   $form['maincntr']['subcontainer']['somewrapper']['some_button'] = [
   *     '#type' => 'submit',
   *     '#value' => $this->t('Do something'),
   *     '#name' => 'do_something',
   *     '#ajax' => [
   *       'callback' => [
   *         get_class($this),
   *         'buildAjaxParentSubForm',
   *       ],
   *       'wrapper' => $form['maincntr']['subcontainer']['#attributes']['id'],
   *       'method' => 'replace',
   *       'effect' => 'fade',
   *     ],
   *   ];
   * @endcode
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The part of the form to return as AJAX.
   */
  public static function buildAjaxParentSubForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $triggering_element = $form_state->getTriggeringElement();
    $parents = $triggering_element['#array_parents'];

    // Try to retrieve the container parent.
    $container = $form;
    while (!empty($parents)) {
      $parent = array_shift($parents);
      $container = $container[$parent];
      if ($container['#attributes']['id'] == $triggering_element['#ajax']['wrapper']) {
        return $container;
      }
    }

    // Otherwise, fallback to triggering element's parent.
    $parents = array_slice($triggering_element['#array_parents'], 0, -1);
    return NestedArray::getValue($form, $parents);
  }

  /**
   * Handles AJAX subform update for grandparent.
   *
   * It works based on a triggering element that will only update its
   * grandparent form element.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The part of the form to return as AJAX.
   */
  public static function buildAjaxGrandParentSubForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $triggering_element = $form_state->getTriggeringElement();

    // Retrieve the parents, so we can climb back up from 2 levels the tree and
    // do not have to hard code the unknown position of our subform.
    $parents = array_slice($triggering_element['#array_parents'], 0, -2);

    // Use NestedArray to get the element in $form at the path that $parents
    // describes.
    return NestedArray::getValue($form, $parents);
  }

  /**
   * Form validation handler.
   *
   * @param array $form
   *   An associative array containing the structure of the plugin form as built
   *   by static::buildConfigurationForm().
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   *
   * @see \Drupal\Core\Plugin\PluginFormInterface::validateConfigurationForm()
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
  }

  /**
   * Form submission handler.
   *
   * @param array $form
   *   An associative array containing the structure of the plugin form as built
   *   by static::buildConfigurationForm().
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   *
   * @see \Drupal\Core\Plugin\PluginFormInterface::submitConfigurationForm()
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    if ($this instanceof ConfigurableInterface) {
      $this->setConfiguration($form_state->getValues());
    }
  }

}
